/*
   Copyright (C) 2004 Michael Nelson Moran

   This file is part of OSCL.

   OSCL is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   OSCL is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   You should have received a copy of the GNU General Public License
   along with OSCL; see the file COPYING.  If not, write to the Free
   Software Foundation, 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To negotiate other licensing arrangments/agreements contact
   the copyright holder:

   Michael N. Moran
   5009 Old Field Ct.         mike@mnmoran.org
   Kennesaw, GA, USA 30144    http://mnmoran.org
*/

#ifndef _oscl_bignum_valueh_
#define _oscl_bignum_valueh_
#include "oscl/compiler/types.h"

/** */
namespace Oscl {
/** */
namespace BigNum {

#if 1
// real    0m7.722s
// user    0m7.320s
// sys     0m0.010s
//
// real    0m7.551s
// user    0m7.320s
// sys     0m0.010s
//
// real    0m7.782s
// user    0m7.350s
// sys     0m0.010s
//
// real    0m8.530s
// user    0m7.500s
// sys     0m0.020s
//
// real    0m7.635s
// user    0m7.320s
// sys     0m0.000s

typedef Unsigned16	FragType;
typedef Unsigned32	DoubleFragType;
#else
// real    0m10.584s
// user    0m10.180s
// sys     0m0.010s
//
// real    0m10.744s
// user    0m10.170s
// sys     0m0.010s
//
// real    0m10.390s
// user    0m10.170s
// sys     0m0.010s
typedef Unsigned32	FragType;
typedef Unsigned64	DoubleFragType;
#endif

enum{nBytesPerFrag=(sizeof(FragType))};
enum{nFragBits=(nBytesPerFrag*8)};

/** */
template <unsigned size>
class Value {
		/** */
		friend class Value<size*2>;
		/** */
		friend class Value<size/2>;
	private:
		/** */
		FragType	_value[size];
	public:
		/** */
		Value() throw();
		/** */
		Value(	const unsigned char	value[],
				unsigned			length
				) throw();
		/** */
		Value(const FragType value[size]) throw();
		/** */
		Value(const Value& value) throw();
		/** */
		Value(const Value<size/2>& value) throw();
		/** */
		explicit Value(const Value<size*2>& value) throw();
		// Assignment
		/** */
		Value&	operator = (const Value& value) throw();
		/** */
		Value&	operator = (unsigned value) throw();
		// Shift
		/** */
		Value&	operator <<= (unsigned value) throw();
		/** */
		Value&	operator >>= (unsigned value) throw();
		/** */
		Value	operator << (unsigned value) throw();
		/** */
		Value	operator >> (unsigned value) throw();
		// Comparison
		/** */
		bool	operator < (const Value& other) const throw();
		/** */
		bool	operator <= (const Value& other) const throw();
		/** */
		bool	operator > (const Value& other) const throw();
		/** */
		bool	operator >= (const Value& other) const throw();
		/** */
		bool	operator == (const Value& other) const throw();
		/** */
		bool	operator != (const Value& other) const throw();
		// Arithmetic
		/** */
		Value	operator ~ () const throw();
		/** Prefix increment */
		Value	operator ++ () throw();
		/** Prefix decrement */
		Value	operator -- () throw();
		/** Postfix increment */
		Value	operator ++ (int) throw();
		/** Postfix decrement */
		Value	operator -- (int) throw();
		/** Change sign */
		Value	operator - () throw();
		/** */
		Value&	operator += (const Value& value) throw();
		/** */
		Value&	operator -= (const Value& value) throw();
		/** */
		Value&	operator %= (const Value<size/2>& value) throw();
		/** */
		Value&	operator *= (const Value<size/2> value) throw();
		/** */
		Value&	powXmodY(const Value& pow,const Value& mod) throw();
		/** */
		Value&	pow(const Value& pow) throw();
		/** */
		Value&	squareModY(const Value& mod) throw();
		/** */
		Value<size*2>	square() throw();
		/** */
		bool	msb() const throw();
		/** */
		bool	lsb() const throw();
		/** */
		unsigned	ms1Bit() const throw();
	};

template <unsigned size>
Value<size>::Value() throw()
		{
	}

template <unsigned size>
Value<size>::Value(	const unsigned char value[],
					unsigned			length
					) throw()
		{
	unsigned	nFrags	= length/nBytesPerFrag;
	unsigned	nExtra	= length%nBytesPerFrag;
	if(nFrags>size){
		nFrags	= size;
		nExtra	= 0;
		}
	const unsigned char*	p	= &value[length-1];
	FragType	frag=0;
	for(unsigned i=0;i<nFrags;++i){
		p	-= nBytesPerFrag;
		for(unsigned j=0;j<nBytesPerFrag;++j){
			frag	<<= 8;
			frag	|= p[j+1];
			}
		_value[(size-1)-i]	= frag;
		}
	if(nExtra){
		frag=0;
		for(unsigned i=0;i<nExtra;++i,--p){
			frag	|= *p;
			frag	<<= 8;
			}
		_value[(size-1)-nFrags]	= frag;
		}
	unsigned	nZeroFill	= size-nFrags;
	if(nZeroFill && nExtra){
		nZeroFill	-= 1;
		}
	for(unsigned i=0;i<nZeroFill;++i){
		_value[i]	= 0;
		}
	}

template <unsigned size>
Value<size>::Value(const FragType value[size]) throw()
		{
	for(unsigned i=0;i<size;++i){
		_value[i]	= value[i];
		}
	}

template <unsigned size>
Value<size>::Value(const Value<size/2>& value) throw()
		{
	FragType*	p	= _value;
	for(unsigned i=0;i<(size/2);++i,++p){
		*p	= 0;
		}
	for(unsigned i=0;i<(size/2);++i,++p){
		*p	= value._value[i];
		}
	}

template <unsigned size>
Value<size>::Value(const Value<size*2>& value) throw()
		{
	const FragType*	p	= &value._value[size];
	for(unsigned i=0;i<size;++i,++p){
		_value[i]	= *p;
		}
	}

template <unsigned size>
Value<size>::Value(const Value& value) throw()
		{
	for(unsigned i=0;i<size;++i){
		_value[i]	= value._value[i];
		}
	}

template <unsigned size>
Value<size>&	Value<size>::operator = (const Value<size>& value) throw()
		{
	for(unsigned i=0;i<size;++i){
		_value[i]	= value._value[i];
		}
	return *this;
	}

//
// value=19
// nFrags	= 19/8 = 2
// nBits	= 19%8 = 3
// nMove	= size-nFrags = 6-2-1 = 3
// size		= 6 bytes;
//																						src
//										src[0]			src[1]
//         0               1               2               3               4               5
// |-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|
//  a b c d e f g h i j k l m n o p q s t u v w x y z A B C D E F G H J K L M N O P Q S T U V W X Y 
// |-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|
//  u v w x y z A B C D E F G H J K L M N O P Q S T U V W X Y 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
// |-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|
//														dest
//		dest

template <unsigned size>
inline Value<size>&	Value<size>::operator <<= (unsigned value) throw() {
	unsigned	nFrags	= value/nFragBits;
	unsigned	nBits	= value%nFragBits;
	if(nFrags >= size){
		*this	= 0;
		return *this;
		}
	unsigned			nMove	= size-nFrags-1;
	FragType*		src	= &_value[nFrags];
	FragType*		dest= &_value[0];
	for(unsigned i=0;i<nMove;++i){
		DoubleFragType	source	= src[0];
		source	<<= nFragBits;
		source	|=	src[1];
		source	>>= (nFragBits-nBits);
		*dest	= (FragType)(source);
		++src;
		++dest;
		}
	*dest	= src[0];
	*dest	<<= nBits;
	++dest;
	for(unsigned i=0;i<nFrags;++i){
		*dest	= 0;
		++dest;
		}
	return *this;
	}

template <unsigned size>
inline Value<size>	Value<size>::operator << (unsigned value) throw() {
	Value<size>	result(*this);
	result <<= value;
	return result;
	}

//
// value=19
// nFrags	= 19/nFragBits = 2
// nBits	= 19%nFragBits = 3
// nMove	= size-nFrags-1	= 6-2-1 = 3
// size		= 6 bytes;
//										src[0]			src[1]
//		src[0]			src[1]
//		src[1]
//         0               1               2               3               4               5
// |-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|
//  a b c d e f g h i j k l m n o p q s t u v w x y z A B C D E F G H J K L M N O P Q S T U V W X Y 
// |-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|
//																						dest
//														dest
//										dest
// |-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|
//  0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 a b c d e f g h i j k l m n o p q s t u v w x y z A B C D
// |-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|
//
template <unsigned size>
inline Value<size>&	Value<size>::operator >>= (unsigned value) throw() {
	unsigned	nFrags	= value/nFragBits;
	unsigned	nBits	= value%nFragBits;
	if(nFrags >= size){
		*this	= 0;
		return *this;
		}
	unsigned			nMove	= size-nFrags-1;
	FragType*		dest	= &_value[(size-1)];
	FragType*		src		= &_value[(size-1)-nFrags-1];
	for(unsigned i=0;i<(nMove);++i){
		DoubleFragType	source	= src[0];
		source	<<=	nFragBits;
		source	|=	src[1];
		*dest	= (FragType)((source >> nBits));
		--src;
		--dest;
		}
	*dest	= src[1] >> nBits;
	dest	= &_value[0];
	for(unsigned i=0;i<nFrags;++i){
		*dest	= 0;
		++dest;
		}
	return *this;
	}

template <unsigned size>
inline Value<size>	Value<size>::operator >> (unsigned value) throw() {
	Value<size>	result(*this);
	result >>= value;
	return result;
	}

//
// sizeof(FragType) == 1
// sizeof(value)	== 4
//	nValuesInFragType	== 4
//	nFragTypesInValue	== 0
//
// sizeof(FragType) == 4
// sizeof(value)	== 4
//	nValuesInFragType	== 1
//	nFragTypesInValue	== 1
//
template <unsigned size>
Value<size>&	Value<size>::operator = (unsigned value) throw() {
	unsigned	nFragTypesInValue	= sizeof(value)/sizeof(FragType);
//	unsigned	nValuesInFragType	= sizeof(FragType)/sizeof(value);
	for(unsigned i=0;i<((size)-nFragTypesInValue);++i){
		_value[i]	= 0;
		}
	for(unsigned i=0;i<(nFragTypesInValue);++i){
		_value[(size-1)-i]	= value>>(i*nFragBits);
		}
	return *this;
	}

template <unsigned size>
inline Value<size>	Value<size>::operator ~ () const throw(){
	Value<size>	result;
	for(unsigned i=0;i<size;++i){
		result._value[i]	= ~_value[i];
		}
	return result;
	}

template <unsigned size>
Value<size>	Value<size>::operator ++ () throw(){
	DoubleFragType	carry	= 1;
	for(signed i=(size-1);i>=0;--i){
		carry		+= _value[i];
		_value[i]	= (FragType)carry;
		carry	>>= nFragBits;
		}
	return *this;
	}

template <unsigned size>
inline Value<size>	Value<size>::operator ++ (int) throw(){
	Value<size>	result(*this);
	++*this;
	return result;
	}

template <unsigned size>
Value<size>	Value<size>::operator -- () throw(){
	DoubleFragType	carry	= 0;
	for(signed i=(size-1);i>=0;--i){
		carry		+= _value[i] + (~((~((DoubleFragType)0))<<nFragBits));
		_value[i]	= (FragType)carry;
		carry	>>= nFragBits;
		}
	return *this;
	}

template <unsigned size>
inline Value<size>	Value<size>::operator -- (int) throw(){
	Value<size>	result(*this);
	--*this;
	return result;
	}

template <unsigned size>
inline Value<size>	Value<size>::operator - () throw(){
	Value<size>	result(*this);
	result	= ~result;
	++result;
	return result;
	}

template <unsigned size>
Value<size>&	Value<size>::operator += (const Value<size>& value) throw(){
	DoubleFragType	carry	= 0;
	for(signed i=(size-1);i>=0;--i){
		carry		+= _value[i];
		carry		+= value._value[i];
		_value[i]	= (FragType)carry;
		carry		>>=	nFragBits;
		}
	return *this;
	}

template <unsigned size>
Value<size>&	Value<size>::operator -= (const Value<size>& value) throw(){
	DoubleFragType	carry	= 1;
	for(signed i=(size-1);i>=0;--i){
		carry		+= ((FragType)(~value._value[i]));
		carry		+= _value[i];
		_value[i]	= (FragType)carry;
		carry		>>=	nFragBits;
		}
	return *this;
	}

template <unsigned size>
unsigned	Value<size>::ms1Bit() const throw(){
	// The value must NOT be zero
	for(unsigned i=0;i<size;++i){
		if(_value[i]){
			for(unsigned j=(nFragBits-1);j != ((unsigned)~0);--j){
				if(_value[i] & (1<<j)){
					return (((size-1)-i)*nFragBits)+j;
					}
				}
			}
		}
	return ((unsigned)~0);	// Only if the value is zero
	}

template <unsigned size>
inline bool	Value<size>::operator < (const Value& other) const throw(){
	for(unsigned i=0;i<size;++i){
		if(other._value[i] != _value[i]){
			return (_value[i] < other._value[i]);
			}
		}
	return false;
	}

template <unsigned size>
inline bool	Value<size>::operator <= (const Value& other) const throw(){
	for(unsigned i=0;i<size;++i){
		if(other._value[i] != _value[i]){
			return (_value[i] < other._value[i]);
			}
		}
	return true;
	}

template <unsigned size>
inline bool	Value<size>::operator > (const Value& other) const throw(){
	for(unsigned i=0;i<size;++i){
		if(other._value[i] != _value[i]){
			return (_value[i] > other._value[i]);
			}
		}
	return false;
	}

template <unsigned size>
inline bool	Value<size>::operator >= (const Value& other) const throw(){
	for(unsigned i=0;i<size;++i){
		if(other._value[i] != _value[i]){
			return (_value[i] > other._value[i]);
			}
		}
	return true;
	}

template <unsigned size>
bool	Value<size>::operator == (const Value& other) const throw(){
	for(unsigned i=0;i<size;++i){
		if(other._value[i] != _value[i]){
			return false;
			}
		}
	return true;
	}

template <unsigned size>
bool	Value<size>::operator != (const Value& other) const throw(){
	for(unsigned i=0;i<size;++i){
		if(other._value[i] != _value[i]){
			return true;
			}
		}
	return false;
	}

template <unsigned size>
Value<size>&	Value<size>::operator %= (const Value<size/2>& value) throw(){
	Value<size>	modulus	= value;
	unsigned	msbOfModulus	= modulus.ms1Bit();
	// Presumably, the modulus is NOT zero
	unsigned	msbOfDividend	= ms1Bit();
	if(msbOfDividend == ((unsigned)~0)){
		// The dividend is zero!
		// 0%anything == zero remainder
		return *this;
		}
	if(msbOfModulus > msbOfDividend){
		// This value *is* the remainder.
		return *this;
		}
	unsigned	nShifts = msbOfDividend - msbOfModulus;
	modulus	<<= nShifts;
	unsigned	nIterations = nShifts + 1;
	for(unsigned i=0;i<nIterations;++i){
		if(*this < modulus){
			modulus	>>=1;
			continue;
			}
		else {
			*this	-= modulus;
			modulus	>>=1;
			}
		}
	return *this;
	}

//	size	= 64
//	size/2	= 32
//	last64	= 63
//	last32	= 31
//			3				2				1
//			V				V				V
//										 last32											last64
//         0               1               2               3               4               5
// |-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|
//  a b c d e f g h i j k l m n o p q s t u v w x y z A B C D E F G H J K L M N O P Q S T U V W X Y 
// |-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|
//  u v w x y z A B C D E F G H J K L M N O P Q S T U V W X Y 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
// |-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|
template <unsigned size>
Value<size>&	Value<size>::operator *= (const Value<size/2> value) throw(){
	Value<size>	multiplicand(*this);
	*this	= 0;
	for(unsigned i=((size/2)-1);i!=((unsigned)~0);--i){
		Value<size>		temp;
		temp	= 0;
		DoubleFragType		carry	= 0;
		unsigned	shift	= (((size/2)-1)-i);
		FragType	multiplier	= value._value[i];
		for(unsigned j=((size)-1);j>=(size/2-1);--j){
			DoubleFragType	acc;
			acc	= multiplicand._value[j];
			acc	*= multiplier;
			acc	+= carry;
			temp._value[j-shift]	= (FragType)acc;
			carry	= acc;
			carry	>>= nFragBits;
			}
		*this			+= temp;
		}
	return *this;
	}

//template <unsigned size>
//Value<size>	Value<size>::operator * (const Value<size/2>& value) throw(){
//	Value<size>	result	= *this;
//	result	*= value;
//	return result;
//	}

template <unsigned size>
inline bool	Value<size>::msb() const throw(){
	return (_value[0] & (1<<(nFragBits-1)));
	}

template <unsigned size>
inline bool	Value<size>::lsb() const throw(){
	return (_value[(size-1)] & 0x01);
	}

template <unsigned size>
Value<size>&	Value<size>::powXmodY(	const Value&	pow,
										const Value&	mod
										) throw(){
	Value<size>			exponent(pow);
	Value<size>			accumulator(*this);
	Value<size*2>		result = *this;
	const unsigned		nPowBytes	= (pow.ms1Bit()+(nFragBits-1))/nFragBits;
	result	= 1;
	for(unsigned i=0;i<nPowBytes;++i){
		for(unsigned j=0;j<nFragBits;++j){
			if(exponent._value[(size-1)-i] & (1<<j)){
				result	*= accumulator;
				result	%= mod;
				}
			accumulator.squareModY(mod);
			}
		}
	*this	= Value<size>(result);
	return *this;
	}

template <unsigned size>
Value<size>&	Value<size>::pow(const Value& pow) throw(){
	Value<size>				accumulator(*this);
	*this	= 1;
	for(unsigned i=0;i<size;++i){
		for(unsigned j=0;j<nFragBits;++j){
			if(pow._value[(size-1)-i] & (1<<j)){
				*this	*= accumulator;
				}
			accumulator.square();
			}
		}
	return *this;
	}

template <unsigned size>
Value<size>&	Value<size>::squareModY(const Value& mod) throw(){
	Value<size*2>	result(*this);
	result	*= *this;
	result	%= mod;
	*this	= Value<size>(result);
	return *this;
	}

template <unsigned size>
inline Value<size*2>	Value<size>::square() throw(){
	Value<size*2>	result(*this);
	result	*= *this;
	return result;
	}

}
}

#endif
