/*
   Copyright (C) 2004 Michael Nelson Moran

   This file is part of OSCL.

   OSCL is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   OSCL is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   You should have received a copy of the GNU General Public License
   along with OSCL; see the file COPYING.  If not, write to the Free
   Software Foundation, 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To negotiate other licensing arrangments/agreements contact
   the copyright holder:

   Michael N. Moran
   5009 Old Field Ct.         mike@mnmoran.org
   Kennesaw, GA, USA 30144    http://mnmoran.org
*/

#ifndef _oscl_bits_mutexedfieldh_
#define _oscl_bits_mutexedfieldh_
#include "oscl/bits/fieldapi.h"
#include "oscl/bits/bitfield.h"
#include "oscl/mt/mutex/simple.h"
#include "oscl/mt/scopesync.h"

/** */
namespace Oscl {
/** */
namespace Bits {

/** */
template <class Type>
class MutexedField : public FieldApi<Type> {
	private:
		/** */
		BitField<Type>					_field;

		/** */
		Oscl::Mt::Mutex::Simple::Api&	_mutex;

	public:
		/** Constructor
		 */
		MutexedField(Mt::Mutex::Simple::Api& mutex,Type initialValue=0) throw();

		/** Change the bits, specified as a mask of affectedBits,
			to the values specified in the bitValues mask.
		 */
		void	changeBits(	Type	affectedBits,
							Type	bitValues
							) throw();

		/** Clear the bits specified as a mask of affectedBits.
		 */
		void	clearBits(Type affectedBits) throw();

		/** Set the bits specified as a mask of affectedBits.
		 */
		void	setBits(Type affectedBits) throw();

		/** Set the bits specified as a mask of affectedBits
			if all of the affected bits are zero, otherwise
			the bits are unchanged. Returns true if the
			affected bits were changed to ones by this invocation 
			of the operation, and false if any of the bits were
			set to one prior to its invocation. This operation
			is atomic to the degree provided by the Mutex.
		 */
		bool	testAndSet(Type affectedBits) throw();

		/** Compare the interestingBits of the port with the
			corresponding bits of the bitValues mask.
		 */
		bool	equal(Type interestingBits,Type bitValues) const throw();

		/** Explicitly read the whole value of the bit field.
			Operation is not necessarily const if this field
			represents a register which has side effects when read.
		 */
		Type	read() const throw();
	};

template <class Type>
MutexedField<Type>::MutexedField(	Mt::Mutex::Simple::Api&	mutex,
									Type					initialValue
									) throw():
		_mutex(mutex)
		{
	_field	= initialValue;
	}

template <class Type>
void	MutexedField<Type>::changeBits(	Type	affectedBits,
										Type	bitValues
										) throw(){
	Mt::ScopeSync	ss(_mutex);
	_field.changeBits(affectedBits,bitValues);
	}

template <class Type>
void	MutexedField<Type>::clearBits(Type affectedBits) throw(){
	Mt::ScopeSync	ss(_mutex);
	_field.clearBits(affectedBits);
	}

template <class Type>
void	MutexedField<Type>::setBits(Type affectedBits) throw(){
	Mt::ScopeSync	ss(_mutex);
	_field.setBits(affectedBits);
	}

template <class Type>
bool	MutexedField<Type>::testAndSet(Type affectedBits) throw(){
	Mt::ScopeSync	ss(_mutex);
	return _field.testAndSet(affectedBits);
	}

template <class Type>
bool	MutexedField<Type>::equal(	Type	interestingBits,
									Type	bitValues
									) const throw(){
	return _field.equal(interestingBits,bitValues);
	}

template <class Type>
Type	MutexedField<Type>::read() const throw(){
	return _field.read();
	}

}
}

#endif
