/*
   Copyright (C) 2004 Michael Nelson Moran

   This file is part of OSCL.

   OSCL is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   OSCL is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   You should have received a copy of the GNU General Public License
   along with OSCL; see the file COPYING.  If not, write to the Free
   Software Foundation, 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To negotiate other licensing arrangments/agreements contact
   the copyright holder:

   Michael N. Moran
   5009 Old Field Ct.         mike@mnmoran.org
   Kennesaw, GA, USA 30144    http://mnmoran.org
*/

#include <new>
#include "oscl/bits/bitfield.h"
#include "driver.h"
#include "oscl/cpu/sync.h"
#include "oscl/cache/operations.h"
#include "oscl/error/fatal.h"

extern void UsbPciConfigDump() throw();

using namespace Oscl::Usb::Ohci::HCD;

Driver::Driver(	Oscl::Usb::Ohci::Map&			reg,
				Oscl::Usb::Ohci::Hcca&			hcca,
				TransactionDescMem*				tdArray,
				unsigned						tdArraySize,
				EndpointDescMem*				edArray,
				unsigned						edArraySize,
				Oscl::Bus::Api&					dmaBus,
				Oscl::Mt::Itc::ServerApi&		server,
				PipeMem*						pipeMem,
				unsigned						maxEndpoints
				) throw():
		Oscl::Mt::Itc::Srv::CloseSync(*this,server.getPostMsgApi()),
		_enumMonitorVar(	server.getPostMsgApi(),
							*this,
							*this,
							*this,
							*this,
							*this,
							*this
							),
		_bwAlloc(hcca.interruptTable),
		_reg(reg),
		_control(reg.controlStatus.control),
		_cmdStatus(reg.controlStatus.commandStatus),
		_irqStatus(reg.controlStatus.interruptStatus),
		_irqEnable(reg.controlStatus.interruptEnable),
		_irqDisable(reg.controlStatus.interruptDisable),
		_hcca(hcca),
		_dmaBus(dmaBus),
		_server(server),
		_open(false),
		_rhOpenReq(0),
		_rhCloseReq(0),
		_rhChangeReq(0),
		_rhSAP(*this,server.getPostMsgApi()),
		_hubPortSAP(*this,server.getPostMsgApi()),
		_msgPipeAllocSAP(*this,server.getPostMsgApi()),
		_intInPipeAllocSAP(*this,server.getPostMsgApi()),
		_bulkInPipeAllocSAP(*this,server.getPostMsgApi()),
		_bulkOutPipeAllocSAP(*this,server.getPostMsgApi()),
		_addressZeroLocked(false),
		_lsDefaultPipe(),
		_hsDefaultPipe(),
		_unexpectedTransfer(*this,&HCD::Driver::unexpectedDone),
		_controlListDisableCount(0),
		_periodicListDisableCount(0),
		_bulkListDisableCount(0),
		_pipeMem(pipeMem)
		{
	for(unsigned i=0;i<maxUsbDeviceAddresses;++i){
		_allocatedUsbAddr[i]	= false;
		}
	for(unsigned i=0;i<tdArraySize;++i){
		_freeTDList.put(&tdArray[i]);
		}
	for(unsigned i=0;i<edArraySize;++i){
		_freeEDList.put(&edArray[i]);
		}
	for(unsigned i=0;i<maxEndpoints;++i){
		_freePipeMem.put(&_pipeMem[i]);
		}
	for(unsigned i=0;i<nDeferredTasks;++i){
		_freeDtMem.put(&_dtMem[i]);
		}
	unsigned long	hccax;
	_dmaBus.cpuToBus(&_hcca,hccax);
	_reg.memoryPointer.hcca	= hccax;
	_lsDefaultPipe	= allocMessagePipe(	Address(0),
										EndpointID(0),
										8,
										true
										);
	_hsDefaultPipe	= allocMessagePipe(	Address(0),
										EndpointID(0),
										8,
										false
										);

	// Create dummy ED for bulk list.
	EndpointDescMem*	edMem	= allocEdMem();
	if(!edMem){
		// This should NEVER happen
		for(;;);
		}
	_bulkEpListHead	= new (edMem)
			Oscl::Usb::Ohci::ED::Descriptor(	getDmaBusApi(),
												0,0,0,false,false
												);
	_bulkEpListHead->skip();

	// Create control pipe list
	_controlEpListHead	= &_lsDefaultPipe->_pipe.getED();
	_controlEpListHead->insert(_hsDefaultPipe->_pipe.getED());
	_enumDcpVar	= new(&_enumDcpVarMem)
					Oscl::Usb::HCD::Enum::
					DefCntrlPipe::Var(	server.getPostMsgApi(),
										*this,
										*this,
										_lsDefaultPipe->_pipe,
										_hsDefaultPipe->_pipe
										);
	server.setSignaledApi(this);
	}
Driver::~Driver(){
	_server.setSignaledApi(0);
	}

RootHub::Req::Api::SAP&		Driver::getRootHubSAP() throw(){
	return _rhSAP;
	}

Oscl::Usb::Pipe::Message::Req::Api::SAP&	Driver::getLsEpSAP() throw(){
	return _lsDefaultPipe->_pipe.getSAP();
	}

Oscl::Usb::Pipe::Message::Req::Api::SAP&	Driver::getHsEpSAP() throw(){
	return _hsDefaultPipe->_pipe.getSAP();
	}

Oscl::Usb::Pipe::Message::SyncApi&		Driver::getLsSyncApi() throw(){
	return _lsDefaultPipe->_pipe.getSyncApi();
	}

Oscl::Usb::Pipe::Message::SyncApi&		Driver::getHsSyncApi() throw(){
	return _hsDefaultPipe->_pipe.getSyncApi();
	}

Oscl::Usb::Alloc::Addr::Req::Api::SAP&	Driver::getHubPortSAP() throw(){
	return _hubPortSAP;
	}

Oscl::Usb::Alloc::Pipe::
Message::Req::Api::SAP&		Driver::getMsgPipeAllocSAP() throw(){
	return _msgPipeAllocSAP;
	}

Oscl::Usb::Alloc::Pipe::
Interrupt::IN::Req::Api::SAP&	Driver::getInterruptInPipeAllocSAP() throw(){
	return _intInPipeAllocSAP;
	}

Oscl::Usb::Alloc::Pipe::
Bulk::IN::Req::Api::SAP&	Driver::getBulkInPipeAllocSAP() throw(){
	return _bulkInPipeAllocSAP;
	}

Oscl::Usb::Alloc::Pipe::
Bulk::OUT::Req::Api::SAP&	Driver::getBulkOutPipeAllocSAP() throw(){
	return _bulkOutPipeAllocSAP;
	}

Oscl::Usb::HCD::Enum::
DefCntrlPipe::Api&				Driver::getEnumDefCntrlPipeApi() throw(){
	return *_enumDcpVar;
	}

Oscl::Usb::HCD::Enum::
Monitor::Api&					Driver::getEnumMonitorApi() throw(){
	return _enumMonitorVar;
	}

bool	Driver::usbAddressAvailable() throw(){
	for(unsigned i=1;i<maxUsbDeviceAddresses;++i){
		if(!_allocatedUsbAddr[i]){
			return true;
			}
		}
	return false;
	}

void	Driver::insertToControlList(Oscl::Usb::Ohci::ED::Message& ed) throw(){
	_controlEpListHead->insert(ed);
	}

void	Driver::removeFromControlList(Oscl::Usb::Ohci::ED::Message& ed) throw(){
	// Takes advantage of the fact that the list is never empty.
//	_controlEpListHead->remove(&ed);
	Oscl::Usb::Ohci::ED::Descriptor*	next;
	for(	next=_controlEpListHead;
			next;
			next	= next->getNextEndpointDescriptor()
			){
		if(next->getNextEndpointDescriptor() == &ed){
			next->setNextED(ed.getNextEndpointDescriptor());
			return;
			}
		}
	}

void	Driver::insertToBulkList(Oscl::Usb::Ohci::ED::Descriptor& ed) throw(){
	_bulkEpListHead->insert(ed);
	}

void	Driver::removeFromBulkList(Oscl::Usb::Ohci::ED::Descriptor& ed) throw(){
	// Takes advantage of the fact that the list is never empty.
	Oscl::Usb::Ohci::ED::Descriptor*	next;
	for(	next=_bulkEpListHead;
			next;
			next	= next->getNextEndpointDescriptor()
			){
		if(next->getNextEndpointDescriptor() == &ed){
			next->setNextED(ed.getNextEndpointDescriptor());
			return;
			}
		}
	}

Driver::DeferredTaskMem*	Driver::allocDeferredTaskMem() throw(){
	return _freeDtMem.get();
	}

void	Driver::free(DeferredTaskMem* mem) throw(){
	_freeDtMem.put(mem);
	Oscl::Mt::Itc::SrvMsg*	msg	= _pendingDeferredTaskReqList.get();
	if(msg){
		msg->process();
		}
	}

void	Driver::release(ControlListAttachDefTask& dt) throw(){
	dt.~ControlListAttachDefTask();
	free((DeferredTaskMem*)&dt);
	}

void	Driver::release(ControlListDetachDefTask& dt) throw(){
	dt.~ControlListDetachDefTask();
	free((DeferredTaskMem*)&dt);
	}

void	Driver::release(InterruptListAttachDefTask& dt) throw(){
	dt.~InterruptListAttachDefTask();
	free((DeferredTaskMem*)&dt);
	}

void	Driver::release(InterruptListDetachDefTask& dt) throw(){
	dt.~InterruptListDetachDefTask();
	free((DeferredTaskMem*)&dt);
	}

void	Driver::release(BulkInListAttachDefTask& dt) throw(){
	dt.~BulkInListAttachDefTask();
	free((DeferredTaskMem*)&dt);
	}

void	Driver::release(BulkInListDetachDefTask& dt) throw(){
	dt.~BulkInListDetachDefTask();
	free((DeferredTaskMem*)&dt);
	}

void	Driver::release(BulkOutListAttachDefTask& dt) throw(){
	dt.~BulkOutListAttachDefTask();
	free((DeferredTaskMem*)&dt);
	}

void	Driver::release(BulkOutListDetachDefTask& dt) throw(){
	dt.~BulkOutListDetachDefTask();
	free((DeferredTaskMem*)&dt);
	}

MessagePipeRec*	Driver::removeActiveMessageEp(	Oscl::Usb::Pipe::
												Message::PipeApi&	pipe
												) throw(){
	MessagePipeRec*	next;
	for(	next=_activeMessageEpList.first();
			next;
			next=_activeMessageEpList.next(next)
			){
		Oscl::Usb::Pipe::Message::PipeApi*	api	= &next->_pipe;
		if(api == &pipe){
			_activeMessageEpList.remove(next);
			return next;
			}
		}
	return 0;
	}

InterruptInPipeRec*
Driver::removeActiveInterruptInPipe(	Oscl::Usb::Pipe::
										Interrupt::IN::
										PipeApi&			pipe
										) throw(){
	InterruptInPipeRec*	next;
	for(	next=_activeInterruptInPipeList.first();
			next;
			next=_activeInterruptInPipeList.next(next)
			){
		Oscl::Usb::Pipe::Interrupt::IN::PipeApi*	api	= &next->_pipe;
		if(api == &pipe){
			_activeInterruptInPipeList.remove(next);
			return next;
			}
		}
	return 0;
	}

BulkInPipeRec*
Driver::removeActiveBulkInPipe(	Oscl::Usb::Pipe::
								Bulk::IN::
								PipeApi&			pipe
								) throw(){
	BulkInPipeRec*	next;
	for(	next=_activeBulkInPipeList.first();
			next;
			next=_activeBulkInPipeList.next(next)
			){
		Oscl::Usb::Pipe::Bulk::IN::PipeApi*	api	= &next->_pipe;
		if(api == &pipe){
			_activeBulkInPipeList.remove(next);
			return next;
			}
		}
	return 0;
	}

BulkOutPipeRec*
Driver::removeActiveBulkOutPipe(	Oscl::Usb::Pipe::
									Bulk::OUT::
									PipeApi&			pipe
									) throw(){
	BulkOutPipeRec*	next;
	for(	next=_activeBulkOutPipeList.first();
			next;
			next=_activeBulkOutPipeList.next(next)
			){
		Oscl::Usb::Pipe::Bulk::OUT::PipeApi*	api	= &next->_pipe;
		if(api == &pipe){
			_activeBulkOutPipeList.remove(next);
			return next;
			}
		}
	return 0;
	}

void	Driver::disableControlList() throw(){
	if(!_controlListDisableCount){
		_control.disableControlList();
		}
	++_controlListDisableCount;
	}

void	Driver::enableControlList() throw(){
	if(_controlListDisableCount){
		--_controlListDisableCount;
		}
	if(!_controlListDisableCount){
		_control.enableControlList();
		}
	}

void	Driver::disablePeriodicList() throw(){
	if(!_periodicListDisableCount){
		_control.disablePeriodicList();
		}
	++_periodicListDisableCount;
	}

void	Driver::enablePeriodicList() throw(){
	if(_periodicListDisableCount){
		--_periodicListDisableCount;
		}
	if(!_periodicListDisableCount){
		_control.enablePeriodicList();
		}
	}

void	Driver::disableBulkList() throw(){
	if(!_bulkListDisableCount){
		_control.disableBulkList();
		}
	++_bulkListDisableCount;
	}

void	Driver::enableBulkList() throw(){
	if(_bulkListDisableCount){
		--_bulkListDisableCount;
		}
	if(!_bulkListDisableCount){
		_control.enableBulkList();
		}
	}

Oscl::Usb::Address Driver::allocateAddress() throw(){
	for(unsigned i=1;i<maxUsbDeviceAddresses;++i){
		if(!_allocatedUsbAddr[i]){
			_allocatedUsbAddr[i]	= true;
			return i;
			}
		}
	return 0;
	}

void	Driver::freeAddress(Oscl::Usb::Address address) throw(){
	if(!_allocatedUsbAddr[address]){
		Oscl::ErrorFatal::logAndExit("Free of unallocted USB address!\n");
		}
	Oscl::Usb::Alloc::Addr::
	Req::Api::UsbAddrResChangeReq*	next;
	while((next=_usbAddrChangeList.get())){
		next->returnToSender();
		}
	_allocatedUsbAddr[address]	= false;
	}

void	Driver::scheduleOverrun() throw(){
	}

void	Driver::writebackDoneHead(Unsigned32 busTD) throw(){
	TD::Descriptor*				doneStack=0;
	TD::Descriptor*				desc;
	void*						vdesc;

	while(busTD){
		if(!_dmaBus.busToCpu(busTD,vdesc)){
			Oscl::ErrorFatal::logAndExit("no translation\n");
			}
		// FIXME: Need desc=PlatPhysicalToVirtual(desc) here.
		desc	= (TD::Descriptor*)vdesc;
		desc->invalidate();
		busTD	= desc->getNextTD();
		desc->_doneStackLink	= doneStack;
		doneStack				= desc;
		}

	while((desc=doneStack)){
		doneStack	= desc->_doneStackLink;
		desc->getTransferApi().done(*desc);
		}
	Queue<DeferredTaskApi>	deferredTasks	= _tdMemDeferredTaskList;
	DeferredTaskApi*	dt;
	while((dt=deferredTasks.get())){
		dt->process();
		}
	}

void	Driver::startOfFrame() throw(){
	Queue<DeferredTaskApi>	deferredTasks	= _sofDeferredTaskList;
	DeferredTaskApi*	dt;
	while((dt=deferredTasks.get())){
		dt->process();
		}
	if(!_sofDeferredTaskList.first()){
		// List is empty
		_irqDisable.startOfFrame();
		}
	}

void	Driver::resumeDetected() throw(){
	}

void	Driver::unrecoverableError() throw(){
	}

void	Driver::frameNumberOverflow() throw(){
	}

void	Driver::rootHubStatusChange() throw(){
	_irqDisable.rootHubStatusChange();
	if(_rhChangeReq){
		_rhChangeReq->returnToSender();
		_rhChangeReq	= 0;
		}
	}

void	Driver::ownershipChange() throw(){
	}

void	Driver::request(	Oscl::Mt::Itc::Srv::
							Open::Req::Api::OpenReq&	msg
							) throw(){
	_open	= true;

	// FIXME: Need to delay before going operational.
#if 0
//	LINUX equivalents
	_control.controlBulkServiceFourToOne();
	_control.enableIsochronous();
	_control.enablePeriodicList();
#endif

//	_irqEnable.startOfFrame();
	_irqEnable.writebackDone();
	_irqEnable.master();

//	_rootHub.initialize();

	unsigned long	busAddr;
	// Setup control list
	_dmaBus.cpuToBus(_controlEpListHead,busAddr);
	_reg.memoryPointer.
	controlHeadED		= busAddr;

	// Setup bulk list
	_dmaBus.cpuToBus(_bulkEpListHead,busAddr);
	_reg.memoryPointer.
	bulkHeadED		= busAddr;

	Oscl::Usb::Ohci::FmInterval::Reg
	fmInterval	= _reg.frameCounter.fmInterval;
	Unsigned32
	interval	= (		fmInterval
					&	Oscl::Usb::Ohci::FmInterval::FI::FieldMask
					) >> Oscl::Usb::Ohci::FmInterval::FI::Lsb;
	Unsigned32
	largestPacket	= (interval - 220);
	largestPacket	*= 6;
	largestPacket	/= 7;
	fmInterval	^=	Oscl::Usb::Ohci::FmInterval::FIT::FieldMask;
	fmInterval	&=	Oscl::Usb::Ohci::FmInterval::FIT::FieldMask;
	fmInterval	|=	
					(interval << Oscl::Usb::Ohci::FmInterval::FI::Lsb)
				|	(largestPacket << Oscl::Usb::Ohci::FmInterval::FSMPS::Lsb)
				;
	_reg.frameCounter.fmInterval	= fmInterval;

	Unsigned32
	periodicStart	= interval - ((interval*10)/100);
	_reg.frameCounter.periodicStart	= periodicStart;
	OsclCpuInOrderMemoryAccessBarrier();
	_control.enableControlList();
	_control.enableBulkList();
	_control.usbOperational();

	if(_rhOpenReq){
		_rhOpenReq->returnToSender();
		_rhOpenReq	= 0;
		}
	msg.returnToSender();
	}

void	Driver::request(	Oscl::Mt::Itc::Srv::
							Close::Req::Api::CloseReq& msg
							) throw(){
	if(_rhCloseReq){
		_rhCloseReq->returnToSender();
		_rhCloseReq		= 0;
		}
	_irqDisable.master();
	_open	= false;
	msg.returnToSender();
	}

void	Driver::request(	Oscl::Usb::Alloc::Pipe::
							Message::Req::Api::NewReq&	msg
							) throw(){
	DeferredTaskMem*	dtMem	= allocDeferredTaskMem();
	if(!dtMem){
		_pendingDeferredTaskReqList.put(&msg);
		return;
		}
	MessagePipeRec*	ep;
	ep	= allocMessagePipe(	msg.getPayload().getDeviceAddress(),
							msg.getPayload().getEndpointID(),
							msg.getPayload().getMaxPacketSize(),
							msg.getPayload().isLowSpeedDevice()
							);
	if(!ep){
		msg.getPayload().setPipeApi(&ep->_pipe);
		msg.returnToSender();
		free(dtMem);
		return;
		}
	new(dtMem) ControlListAttachDefTask(	*this,
											msg,
											*ep
											);
	_activeMessageEpList.put(ep);
	}

void	Driver::request(	Oscl::Usb::Alloc::Pipe::
							Message::Req::Api::FreeReq&	msg
							) throw(){
	DeferredTaskMem*	dtMem	= allocDeferredTaskMem();
	if(!dtMem){
		_pendingDeferredTaskReqList.put(&msg);
		return;
		}
	new(dtMem) ControlListDetachDefTask(	*this,
											msg
											);
	}

void	Driver::request(	Oscl::Usb::Alloc::Pipe::
							Message::Req::Api::NotifyChangeReq&	msg
							) throw(){
	// if allocMessagePipe() resources are available:
	if(		_freeTDList.first()
		&& _freeEDList.first()
		&& _freePipeMem.first()
		){
		msg.returnToSender();
		return;
		}
	_msgPipeResChangeList.put(&msg);
	}

void	Driver::request(	Oscl::Usb::Alloc::Pipe::
							Message::Req::Api::CancelChangeReq&	msg
							) throw(){
	_msgPipeResChangeList.remove(&msg.getPayload()._toCancel);
	msg.getPayload()._toCancel.returnToSender();
	msg.returnToSender();
	}

void	Driver::request(	Oscl::Usb::Alloc::Pipe::
							Interrupt::IN::Req::Api::NewReq&	msg
							) throw(){
	Oscl::Usb::Alloc::Pipe::
	Interrupt::IN::Req::Api::NewPayload&
	payload	= msg.getPayload();
	Oscl::Usb::Ohci::BW::BwAlloc*
	bwAlloc	= _bwAlloc.find(payload.getBwRec());
	if(!bwAlloc){
		// FIXME Logic Error This should never happen
		while(true);
		}
	InterruptInPipeRec*
	pipe	=	allocInterruptInPipe(	bwAlloc->_ed,
										payload.getDeviceAddress(),
										payload.getEndpointID(),
										payload.getMaxPacketSize(),
										payload.isLowSpeedDevice()
										);
	if(pipe){
		payload.setPipeApi(&pipe->_pipe);
		_activeInterruptInPipeList.put(pipe);
		}
	else{
		payload.setPipeApi(0);
		}
	msg.returnToSender();
	}

void	Driver::request(	Oscl::Usb::Alloc::Pipe::
							Interrupt::IN::Req::Api::FreeReq&	msg
							) throw(){
	Oscl::Usb::Pipe::Interrupt::IN::PipeApi*
	pipeApi	= &msg.getPayload().getPipeApi();
	InterruptInPipeRec*
	pipe	= removeActiveInterruptInPipe(*pipeApi);
	if(!pipe){
		// FIXME: Logic error should never happen
		while(true);
		}
	free(*pipe);
	msg.returnToSender();
	}

void	Driver::request(	Oscl::Usb::Alloc::Pipe::
							Interrupt::IN::Req::Api::NotifyChangeReq&	msg
							) throw(){
	// FIXME
	while(true);
	}

void	Driver::request(	Oscl::Usb::Alloc::Pipe::
							Interrupt::IN::Req::Api::CancelChangeReq&	msg
							) throw(){
	// FIXME
	while(true);
	}

void	Driver::request(	Oscl::Usb::Alloc::Pipe::
							Bulk::IN::Req::Api::NewReq&	msg
							) throw(){
	DeferredTaskMem*	dtMem	= allocDeferredTaskMem();
	if(!dtMem){
		_pendingDeferredTaskReqList.put(&msg);
		return;
		}
	Oscl::Usb::Alloc::Pipe::
	Bulk::IN::Req::Api::NewPayload&
	payload	= msg.getPayload();
	BulkInPipeRec*
	pipe	=	allocBulkInPipe(	payload.getDeviceAddress(),
									payload.getEndpointID(),
									payload.getMaxPacketSize(),
									payload.isLowSpeedDevice()
									);
	if(!pipe){
		msg.getPayload().setPipeApi(&pipe->_pipe);
		msg.returnToSender();
		free(dtMem);
		return;
		}
	new(dtMem) BulkInListAttachDefTask(	*this,
										msg,
										*pipe
										);
	_activeBulkInPipeList.put(pipe);
#if 0
	if(pipe){
		payload.setPipeApi(&pipe->_pipe);
		_activeBulkInPipeList.put(pipe);
		}
	else{
		payload.setPipeApi(0);
		}
	msg.returnToSender();
#endif
	}

void	Driver::request(	Oscl::Usb::Alloc::Pipe::
							Bulk::IN::Req::Api::FreeReq&	msg
							) throw(){
	Oscl::Usb::Pipe::Bulk::IN::PipeApi*
	pipeApi	= &msg.getPayload().getPipeApi();
	BulkInPipeRec*
	pipe	= removeActiveBulkInPipe(*pipeApi);
	if(!pipe){
		// FIXME: Logic error should never happen
		while(true);
		}
	free(*pipe);
	msg.returnToSender();
	}

void	Driver::request(	Oscl::Usb::Alloc::Pipe::
							Bulk::IN::Req::Api::NotifyChangeReq&	msg
							) throw(){
	// FIXME
	while(true);
	}

void	Driver::request(	Oscl::Usb::Alloc::Pipe::
							Bulk::IN::Req::Api::CancelChangeReq&	msg
							) throw(){
	// FIXME
	while(true);
	}

void	Driver::request(	Oscl::Usb::Alloc::Pipe::
							Bulk::OUT::Req::Api::NewReq&	msg
							) throw(){
	DeferredTaskMem*	dtMem	= allocDeferredTaskMem();
	if(!dtMem){
		_pendingDeferredTaskReqList.put(&msg);
		return;
		}
	Oscl::Usb::Alloc::Pipe::
	Bulk::OUT::Req::Api::NewPayload&
	payload	= msg.getPayload();
	BulkOutPipeRec*
	pipe	=	allocBulkOutPipe(	payload.getDeviceAddress(),
									payload.getEndpointID(),
									payload.getMaxPacketSize(),
									payload.isLowSpeedDevice()
									);
	if(!pipe){
		msg.getPayload().setPipeApi(&pipe->_pipe);
		msg.returnToSender();
		free(dtMem);
		return;
		}
	new(dtMem) BulkOutListAttachDefTask(	*this,
											msg,
											*pipe
											);
	_activeBulkOutPipeList.put(pipe);
#if 0
	if(pipe){
		payload.setPipeApi(&pipe->_pipe);
		_activeBulkOutPipeList.put(pipe);
		}
	else{
		payload.setPipeApi(0);
		}
	msg.returnToSender();
#endif
	}

void	Driver::request(	Oscl::Usb::Alloc::Pipe::
							Bulk::OUT::Req::Api::FreeReq&	msg
							) throw(){
	Oscl::Usb::Pipe::Bulk::OUT::PipeApi*
	pipeApi	= &msg.getPayload().getPipeApi();
	BulkOutPipeRec*
	pipe	= removeActiveBulkOutPipe(*pipeApi);
	if(!pipe){
		// FIXME: Logic error should never happen
		while(true);
		}
	free(*pipe);
	msg.returnToSender();
	}

void	Driver::request(	Oscl::Usb::Alloc::Pipe::
							Bulk::OUT::Req::Api::NotifyChangeReq&	msg
							) throw(){
	// FIXME
	while(true);
	}

void	Driver::request(	Oscl::Usb::Alloc::Pipe::
							Bulk::OUT::Req::Api::CancelChangeReq&	msg
							) throw(){
	// FIXME
	while(true);
	}

bool	Driver::interrupt() throw(){
	if(isr()){
		dsr();
		return true;
		}
	return false;
	}

bool	Driver::isr() throw(){
	Unsigned32	doneHead	= _hcca.doneHead;
	if(doneHead && (doneHead & 0x00000001)){
		_irqDisable.master();
		return true;
		}
	else if (_irqStatus.anyPending()){
		_irqDisable.master();
		return true;
		}
	return false;
	}

void	Driver::dsr() throw(){
	if(_open) _server.getSignalApi().suSignal();
	}

void	Driver::initialize() throw(){
	}

void	Driver::signaled() throw(){
	Unsigned32	doneHead	= _hcca.doneHead;
	if(doneHead && !(doneHead & 0x00000001)){	// FIXME: constant
		// In this case, there is no need to read
		// the OHCI registers (delayed PCI read).
		_hcca.doneHead	= 0;
		OsclCpuInOrderMemoryAccessBarrier();
		_irqStatus.ackWritebackDoneHead();
		writebackDoneHead(doneHead);
		_irqEnable.master();
		return;
		}
	_irqStatus.update();
	if(_irqStatus.scheduleOverrun()){
		_irqStatus.ackScheduleOverrun();
		scheduleOverrun();
		}

	if(_irqStatus.writebackDoneHead()){
		OsclCacheDataInvalidateRange(	&_hcca.doneHead,
										sizeof(_hcca.doneHead)
										);
		Unsigned32	doneHead	= _hcca.doneHead;
		_hcca.doneHead			= 0;
		OsclCpuInOrderMemoryAccessBarrier();
		doneHead				&= ~0x00000001;	// FIXME: constant
		_irqStatus.ackWritebackDoneHead();
		writebackDoneHead(doneHead);
		}
	if(_irqStatus.startOfFrame()){
		_irqStatus.ackStartOfFrame();
		startOfFrame();
		}
	if(_irqStatus.resumeDetected()){
		_irqStatus.ackResumeDetected();
		resumeDetected();
		}
	if(_irqStatus.unrecoverableError()){
		_irqStatus.ackUnrecoverableError();
		unrecoverableError();
		}
	if(_irqStatus.frameNumberOverflow()){
		_irqStatus.ackFrameNumberOverflow();
		frameNumberOverflow();
		}
	if(_irqStatus.rootHubStatusChange()){
		_irqStatus.ackRootHubStatusChange();
		rootHubStatusChange();
		}
	if(_irqStatus.ownershipChange()){
		_irqStatus.ackOwnershipChange();
		ownershipChange();
		}

	_irqEnable.master();
	}

void	Driver::request(RootHub::Req::Api::OpenReq& msg) throw(){
	if(_open){
		msg.returnToSender();
		}
	else{
		_rhOpenReq	= &msg;
		}
	}

void	Driver::request(RootHub::Req::Api::CloseReq& msg) throw(){
	if(!_open){
		msg.returnToSender();
		}
	else{
		_rhCloseReq	= &msg;
		}
	}

void	Driver::request(RootHub::Req::Api::ChangeReq& msg) throw(){
	if(!_open) {
		msg.returnToSender();
		return;
		}
	_rhChangeReq	= &msg;
	_irqEnable.rootHubStatusChange();
	}

void	Driver::request(RootHub::Req::Api::CancelAllReq& msg) throw(){
	if(_rhOpenReq) _rhOpenReq->returnToSender();
	if(_rhCloseReq) _rhCloseReq->returnToSender();
	if(_rhChangeReq) _rhChangeReq->returnToSender();
	_rhChangeReq	= 0;
	_rhCloseReq		= 0;
	_rhChangeReq	= 0;
	msg.returnToSender();
	}

void	Driver::request(	Oscl::Usb::Alloc::Addr::
							Req::Api::LockAddressZeroReq&		msg
							) throw(){
	if(_addressZeroLocked){
		_addressZeroLockPendingQ.put(&msg);
		}
	else{
		_addressZeroLocked	= true;
		msg._payload._canceled	= false;
		msg.returnToSender();
		}
	}

void	Driver::request(	Oscl::Usb::Alloc::Addr::
							Req::Api::UnlockAddressZeroReq&	msg
							) throw(){
	using namespace Oscl::Usb::Alloc::Addr;
	_addressZeroLocked	= false;
	msg.returnToSender();
	Req::Api::LockAddressZeroReq*	req	= _addressZeroLockPendingQ.get();
	if(req){
		req->_payload._canceled	= false;
		req->returnToSender();
		}
	}

void	Driver::request(	Oscl::Usb::Alloc::Addr::
							Req::Api::CancelAddrZeroLockReq&	msg
							) throw(){
	
	Oscl::Usb::Alloc::Addr::
	Req::Api::LockAddressZeroReq*
	canceled	= _addressZeroLockPendingQ.remove(&msg.getPayload()._toCancel);
	if(canceled){
		canceled->_payload._canceled	= true;
		canceled->returnToSender();
		}
	msg.returnToSender();
	}

void	Driver::request(	Oscl::Usb::Alloc::Addr::
							Req::Api::AllocateNewAddressReq&	msg
							) throw(){
	Oscl::Usb::Address	usbAddr	= allocateAddress();
	if(usbAddr == 0){
		msg.getPayload().setFailure();
		msg.returnToSender();
		return;
		}
	msg.getPayload().setSuccess(usbAddr);
	msg.returnToSender();
	}

void	Driver::request(	Oscl::Usb::Alloc::Addr::
							Req::Api::FreeAddressReq&		msg
							) throw(){
	freeAddress(msg.getPayload().getAddress());
	msg.returnToSender();
	}

void	Driver::request(	Oscl::Usb::Alloc::Addr::
							Req::Api::UsbAddrResChangeReq&		msg
							) throw(){
	if(usbAddressAvailable()){
		msg.returnToSender();
		}
	else{
		_usbAddrChangeList.put(&msg);
		}
	}

void	Driver::request(	Oscl::Usb::Alloc::Addr::
							Req::Api::CancelUsbAddrResChangeReq&	msg
							) throw(){
	Oscl::Usb::Alloc::Addr::
	Req::Api::UsbAddrResChangeReq*	next;
	next	= _usbAddrChangeList.remove(&msg.getPayload()._toCancel);
	if(next){
		next->returnToSender();
		}
	msg.returnToSender();
	}

//static Oscl::Usb::Alloc::BW::Interrupt::BwRec	bwRec;

/*
	Interrupt IN:
		SYNC		8
		IN PID		8
		Addr		7
		Endpoint	4
		CRC5		5
		EOP			4

		IPD			8

		SYNC		8
		Data0/1 PID	8
		payload		(maxPacketSize*8)
		CRC16		16
		EOP			4

		IPD			8

		SYNC		8
		Ack PID		8
		EOP			3
		--------------
 */

void	Driver::request(	Alloc::BW::Interrupt::
							Req::Api::ReserveReq&	msg
							) throw(){
	DeferredTaskMem*	dtMem	= allocDeferredTaskMem();
	if(!dtMem){
		_pendingDeferredTaskReqList.put(&msg);
		return;
		}
	const unsigned  exemptOverheadBits      = 75;
	const unsigned  nonExemptOverheadBits   = 32;

	ReservePayload&	payload	= msg.getPayload();
	unsigned	nBitTimes;
	nBitTimes	= payload._maxPacketSize * 8;
	nBitTimes	+= nonExemptOverheadBits;
	nBitTimes	*= 7;
	nBitTimes	+= 5;	// round up
	nBitTimes	/= 6;
	nBitTimes	+=	exemptOverheadBits;
	if(!_bwAlloc.bwAvailable(payload._pollPeriodInMilliseconds,nBitTimes)){
		msg.getPayload()._bwRec	= 0;
		msg.returnToSender();
		return;
		}
	new(dtMem) InterruptListAttachDefTask(	*this,
											msg,
											_bwAlloc
											);
	}

void	Driver::request(	Alloc::BW::Interrupt::
							Req::Api::FreeReq&		msg
							) throw(){
	DeferredTaskMem*	dtMem	= allocDeferredTaskMem();
	if(!dtMem){
		_pendingDeferredTaskReqList.put(&msg);
		return;
		}
	new(dtMem) InterruptListDetachDefTask(	*this,
											msg,
											_bwAlloc
											);
	}

void	Driver::request(	Alloc::BW::
							Req::Api::ChangeReq&	msg
							) throw(){
	// FIXME:
	while(true);
	}

void	Driver::request(	Alloc::BW::
							Req::Api::CancelReq&	msg
							) throw(){
	// FIXME:
	while(true);
	}

void	Driver::unexpectedDone(Oscl::Usb::Ohci::TD::Descriptor& desc) throw(){
	Oscl::ErrorFatal::
	logAndExit("Terminating descriptor shouldn't complete\n");
	}

TransactionDescMem*	Driver::allocTdMem() throw(){
	return _freeTDList.get();
	}

EndpointDescMem*	Driver::allocEdMem() throw(){
	return _freeEDList.get();
	}

void				Driver::free(TransactionDescMem* m) throw(){
	_freeTDList.put(m);
	}

void	Driver::free(EndpointDescMem* m) throw(){
	return _freeEDList.put(m);
	}

void				Driver::free(Oscl::Usb::Ohci::TD::Descriptor& d) throw(){
	d.~Descriptor();
	_freeTDList.put((TransactionDescMem*)&d);
	}

void		Driver::queueSofTask(DeferredTaskApi& dt) throw(){
	// List is not empty
	_sofDeferredTaskList.put(&dt);
	_irqEnable.startOfFrame();
	}

void		Driver::cancelSofTask(DeferredTaskApi& dt) throw(){
	if(_sofDeferredTaskList.remove(&dt)){
		if(!_sofDeferredTaskList.first()){
			_irqDisable.startOfFrame();
			}
		}
	}

void		Driver::queueTdMemTask(DeferredTaskApi& dt) throw(){
	// List is not empty
	_tdMemDeferredTaskList.put(&dt);
	}

void		Driver::cancelTdMemTask(DeferredTaskApi& dt) throw(){
	_tdMemDeferredTaskList.remove(&dt);
	}

void		Driver::controlListFilled() throw(){
	_cmdStatus.controlListFilled();
	}

void		Driver::bulkListFilled() throw(){
	_cmdStatus.bulkListFilled();
	}

unsigned    Driver::getCurrentFrameNumber() throw(){
	return _hcca.frameNumber;
	}

Oscl::Bus::Api&	Driver::getDmaBusApi() throw(){
	return _dmaBus;
	}

Oscl::Usb::Ohci::
TD::TransferApi&	Driver::getUnexpectedTransferVector() throw(){
	return _unexpectedTransfer;
	}

MessagePipeRec*
Driver::allocMessagePipe(	Address		usbAddress,
							EndpointID	endpointID,
							unsigned	maxPacketSize,
							bool		lowSpeedDevice
							) throw(){
	// First get all the memory.
	EndpointDescMem*	edMem	= _freeEDList.get();
	TransactionDescMem*	tdMem	= allocTdMem();
	PipeMem*			pipeMem	= _freePipeMem.get();
	if(!edMem || !tdMem || !pipeMem){
		if(edMem) _freeEDList.put(edMem);
		if(tdMem) free(tdMem);
		if(pipeMem) _freePipeMem.put(pipeMem);
		return 0;
		}

	// Next build the endpoint.
	Oscl::Usb::Ohci::ED::Message*	ed;
	TD::Descriptor*					td;
	MessagePipeRec*					ep;
	ed	= new(edMem)ED::Message(	_dmaBus,
									usbAddress,
									endpointID,
									maxPacketSize,
									lowSpeedDevice
									);
	td	= new(tdMem)
				TD::Descriptor(	_unexpectedTransfer,
								_dmaBus
								);

	// Put termination TD on ED list.
	ed->put(td);

	ep	= new(pipeMem) MessagePipeRec(_server.getPostMsgApi(),*ed,*this);

	return ep;
	}

InterruptInPipeRec*
Driver::allocInterruptInPipe(	ED::Descriptor&	ed,
								Address			usbAddress,
								EndpointID		endpointID,
								unsigned		maxPacketSize,
								bool			lowSpeedDevice
								) throw(){
	// First get all the memory.
	PipeMem*			pipeMem	= _freePipeMem.get();
	if(!pipeMem){
		return 0;
		}

	// Get termination TD
	TD::Descriptor*	td		= ed.getTdQueueTailPointer();
	ED::Descriptor*	nextED	= ed.getNextEndpointDescriptor();

	// Next re-build the endpoint.
	ED::GeneralIn*
	ined	= new(&ed)ED::GeneralIn(	_dmaBus,
										usbAddress,
										endpointID,
										maxPacketSize,
										lowSpeedDevice
										);
	InterruptInPipeRec*
	pipe	= new(pipeMem)
				InterruptInPipeRec(	_server.getPostMsgApi(),
									*ined,
									*this
									);
	// Put termination TD on ED list.
	ined->put(td);
	ined->setNextED(nextED);

	return pipe;
	}

BulkInPipeRec*
Driver::allocBulkInPipe(	Address			usbAddress,
							EndpointID		endpointID,
							unsigned		maxPacketSize,
							bool			lowSpeedDevice
							) throw(){
	// First get all the memory.
	EndpointDescMem*	edMem	= _freeEDList.get();
	TransactionDescMem*	tdMem	= allocTdMem();
	PipeMem*			pipeMem	= _freePipeMem.get();
	if(!edMem || !tdMem || !pipeMem){
		if(edMem) _freeEDList.put(edMem);
		if(tdMem) free(tdMem);
		if(pipeMem) _freePipeMem.put(pipeMem);
		return 0;
		}

	// Next build the endpoint.
	Oscl::Usb::Ohci::ED::GeneralIn*	ed;
	TD::Descriptor*					td;
	ed	= new(edMem)ED::GeneralIn(	_dmaBus,
									usbAddress,
									endpointID,
									maxPacketSize,
									lowSpeedDevice
									);
	BulkInPipeRec*
	pipe	= new(pipeMem)
				BulkInPipeRec(	_server.getPostMsgApi(),
								*ed,
								*this
								);

	td	= new(tdMem)
				TD::Descriptor(	_unexpectedTransfer,
								_dmaBus
								);

	// Put termination TD on ED list.
	ed->put(td);

	// Since this operation is invoked during
	// construction (before the bulk list is initialized)
	// we let the caller link the endpoint onto the bulk
	// endpoint list.

	return pipe;
	}

BulkOutPipeRec*
Driver::allocBulkOutPipe(	Address			usbAddress,
							EndpointID		endpointID,
							unsigned		maxPacketSize,
							bool			lowSpeedDevice
							) throw(){
	// First get all the memory.
	EndpointDescMem*	edMem	= _freeEDList.get();
	TransactionDescMem*	tdMem	= allocTdMem();
	PipeMem*			pipeMem	= _freePipeMem.get();
	if(!edMem || !tdMem || !pipeMem){
		if(edMem) _freeEDList.put(edMem);
		if(tdMem) free(tdMem);
		if(pipeMem) _freePipeMem.put(pipeMem);
		return 0;
		}

	// Next build the endpoint.
	TD::Descriptor*					td;
	ED::GeneralOut*
	ed	= new(edMem)ED::GeneralOut(	_dmaBus,
									usbAddress,
									endpointID,
									maxPacketSize,
									lowSpeedDevice
									);
	BulkOutPipeRec*
	pipe	= new(pipeMem)
				BulkOutPipeRec(	_server.getPostMsgApi(),
								*ed,
								*this
								);

	td	= new(tdMem)
				TD::Descriptor(	_unexpectedTransfer,
								_dmaBus
								);

	// Put termination TD on ED list.
	ed->put(td);

	// Since this operation is invoked during
	// construction (before the bulk list is initialized)
	// we let the caller link the endpoint onto the bulk
	// endpoint list.

	return pipe;
	}

#if 0
InterruptOutPipeRec*
Driver::allocInterruptOutPipe(	ED::Descriptor&	ed,
								Address			usbAddress,
								EndpointID		endpointID,
								unsigned		maxPacketSize,
								bool			lowSpeedDevice,
								bool			in
								) throw(){
	// FIXME
	while(true);
	// First get all the memory.
	PipeMem*			pipeMem	= _freePipeMem.get();
	if(!pipeMem){
		return 0;
		}

	// Get termination TD
	TD::Descriptor*	td		= ed.getTdQueueTailPointer();
	ED::Descriptor*	nextED	= ed.getNextEndpointDescriptor();

	new(&ed)ED::GeneralOut(	_dmaBus,
							usbAddress,
							endpointID,
							maxPacketSize,
							lowSpeedDevice
							);
	InterruptOutPipeRec*
	ep	= 0;
//	ep	= new(pipeMem)
//			InterruptOutPipeRec(	_server.getPostMsgApi(),
//									*ed,
//									*this
//									);

	// Put termination TD on ED list.
	ed.put(td);
	ed.setNextED(nextED);

	// Link new ED to control list?

	return ep;
	}
#endif

void	Driver::free(MessagePipeRec& pipe) throw(){
	Oscl::Usb::Ohci::ED::Message&	ed	= pipe._pipe.getED();

	Oscl::Usb::Ohci::TD::Descriptor*
	td	= ed.getTdQueueTailPointer();

	free((TransactionDescMem*)td);
	_freeEDList.put((EndpointDescMem*)&ed);

	pipe.~MessagePipeRec();
	_freePipeMem.put((PipeMem*)&pipe);

	notifyPipeResourceAvailable();
	}

void	Driver::free(InterruptInPipeRec& pipe) throw(){
	Oscl::Usb::Ohci::ED::GeneralIn&	ed	= pipe._pipe.getED();

	Oscl::Usb::Ohci::TD::Descriptor*
	td	= ed.getTdQueueTailPointer();

	free((TransactionDescMem*)td);
	_freeEDList.put((EndpointDescMem*)&ed);

	pipe.~InterruptInPipeRec();
	_freePipeMem.put((PipeMem*)&pipe);

	notifyPipeResourceAvailable();
	}

void	Driver::free(BulkInPipeRec& pipe) throw(){
	Oscl::Usb::Ohci::ED::GeneralIn&	ed	= pipe._pipe.getED();

	Oscl::Usb::Ohci::TD::Descriptor*
	td	= ed.getTdQueueTailPointer();

	free((TransactionDescMem*)td);
	_freeEDList.put((EndpointDescMem*)&ed);

	pipe.~BulkInPipeRec();
	_freePipeMem.put((PipeMem*)&pipe);

	notifyPipeResourceAvailable();
	}

void	Driver::free(BulkOutPipeRec& pipe) throw(){
	Oscl::Usb::Ohci::ED::GeneralOut&	ed	= pipe._pipe.getED();

	Oscl::Usb::Ohci::TD::Descriptor*
	td	= ed.getTdQueueTailPointer();

	free((TransactionDescMem*)td);
	_freeEDList.put((EndpointDescMem*)&ed);

	pipe.~BulkOutPipeRec();
	_freePipeMem.put((PipeMem*)&pipe);

	notifyPipeResourceAvailable();
	}

void	Driver::notifyPipeResourceAvailable() throw(){
	Oscl::Usb::Alloc::Pipe::
	Interrupt::IN::Req::Api::NotifyChangeReq*	iin;
	while((iin=_intInPipeResChangeList.get())){
		iin->returnToSender();
		}

	Oscl::Usb::Alloc::Pipe::Message::Req::Api::NotifyChangeReq*	msg;
	while((msg=_msgPipeResChangeList.get())){
		msg->returnToSender();
		}

	Oscl::Usb::Alloc::Pipe::
	Bulk::IN::Req::Api::NotifyChangeReq*	bulkin;
	while((bulkin=_bulkInPipeResChangeList.get())){
		bulkin->returnToSender();
		}

	Oscl::Usb::Alloc::Pipe::
	Bulk::OUT::Req::Api::NotifyChangeReq*	bulkout;
	while((bulkout=_bulkOutPipeResChangeList.get())){
		bulkout->returnToSender();
		}
	}

//================= ControlListAttachDefTask ========

Driver::ControlListAttachDefTask::
ControlListAttachDefTask(	Driver&				context,
							Oscl::Usb::Alloc::
							Pipe::Message::
							Req::Api::NewReq&	msg,
							MessagePipeRec&		pipe
							) throw():
		_context(context),
		_msg(msg),
		_pipe(pipe)
		{
	_context.disableControlList();
	_currentFrameNumber	= _context.getCurrentFrameNumber();
	queueSofTask();
	}

void	Driver::ControlListAttachDefTask::queueSofTask() throw(){
	DeferredTask<ControlListAttachDefTask>*
	dt	= new(&_dtMem)
			DeferredTask<	ControlListAttachDefTask
							>(	*this,
								&Driver::ControlListAttachDefTask::sofDone
								);
	_context.queueSofTask(*dt);
	}

void	Driver::ControlListAttachDefTask::sofDone() throw(){
	if(_currentFrameNumber == _context.getCurrentFrameNumber()){
		queueSofTask();
		return;
		}
	_msg.getPayload().setPipeApi(&_pipe._pipe);
	_msg.returnToSender();
	_context.insertToControlList(_pipe._pipe.getED());
	_context.enableControlList();
	_context.release(*this);
	}

//================= ControlListDetachDefTask ========

Driver::ControlListDetachDefTask::
ControlListDetachDefTask(	Driver&				context,
							Oscl::Usb::Alloc::
							Pipe::Message::
							Req::Api::FreeReq&	msg
							) throw():
		_context(context),
		_msg(msg)
		{
	_context.disableControlList();
	_currentFrameNumber	= _context.getCurrentFrameNumber();
	queueSofTask();
	}

void	Driver::ControlListDetachDefTask::queueSofTask() throw(){
	DeferredTask<ControlListDetachDefTask>*
	dt	= new(&_dtMem)
			DeferredTask<	ControlListDetachDefTask
							>(	*this,
								&Driver::ControlListDetachDefTask::sofDone
								);
	_context.queueSofTask(*dt);
	}

void	Driver::ControlListDetachDefTask::sofDone() throw(){
	if(_currentFrameNumber == _context.getCurrentFrameNumber()){
		queueSofTask();
		return;
		}
	Oscl::Usb::Pipe::Message::PipeApi*
	pipeApi	= &_msg.getPayload().getPipeApi();
	MessagePipeRec*
	pipe	= _context.removeActiveMessageEp(*pipeApi);
	if(pipe){
		_context.removeFromControlList(pipe->_pipe.getED());
		_context.free(*pipe);
		}
	_msg.returnToSender();
	_context.enableControlList();
	_context.release(*this);
	}

//================= InterruptListAttachDefTask ========

Driver::InterruptListAttachDefTask::
InterruptListAttachDefTask(	Driver&				context,
							Oscl::Usb::Alloc::
							BW::Interrupt::
							Req::Api::
							ReserveReq&			msg,
							Oscl::Usb::Ohci::
							BW::Allocator&		allocator
							) throw():
		_context(context),
		_msg(msg),
		_allocator(allocator)
		{
	_context.disablePeriodicList();
	_currentFrameNumber	= _context.getCurrentFrameNumber();
	queueSofTask();
	}

void	Driver::InterruptListAttachDefTask::queueSofTask() throw(){
	DeferredTask<InterruptListAttachDefTask>*
	dt	= new(&_dtMem)
			DeferredTask<	InterruptListAttachDefTask
							>(	*this,
								&Driver::InterruptListAttachDefTask::sofDone
								);
	_context.queueSofTask(*dt);
	}

void	Driver::InterruptListAttachDefTask::sofDone() throw(){
	if(_currentFrameNumber == _context.getCurrentFrameNumber()){
		queueSofTask();
		return;
		}
	// First get all the memory.
	EndpointDescMem*	edMem	= _context.allocEdMem();
	TransactionDescMem*	tdMem	= _context.allocTdMem();
	if(!edMem || !tdMem ){
		if(edMem) _context.free(edMem);
		if(tdMem) _context.free(tdMem);
		_msg.getPayload()._bwRec	= 0;
		}
	else{
		// Next build the endpoint.
		Oscl::Usb::Ohci::ED::GeneralIn*	ed;
		TD::Descriptor*					td;
		ed	= new(edMem)ED::GeneralIn(	_context.getDmaBusApi(),
										0,		// usbAddress,
										0,		// endpointID,
										0,		// maxPacketSize,
										false	// lowSpeedDevice
										);
		td	= new(tdMem)
					TD::Descriptor(	_context.getUnexpectedTransferVector(),
									_context.getDmaBusApi()
									);
		ed->skip();

		// Put termination TD on ED list.
		ed->put(td);

		Oscl::Usb::Alloc::BW::Interrupt::BwRec*
		bwRec	= _allocator.alloc(*ed);
		if(!bwRec){
			td->~Descriptor();
			_context.free(tdMem);
			ed->~GeneralIn();
			_context.free(edMem);
			}
		_msg.getPayload()._bwRec	= bwRec;
		}
	_msg.returnToSender();
	_context.enablePeriodicList();
	_context.release(*this);
	}

//================= InterruptListDetachDefTask ========

Driver::InterruptListDetachDefTask::
InterruptListDetachDefTask(	Driver&				context,
							Oscl::Usb::Alloc::
							BW::Interrupt::
							Req::Api::
							FreeReq&			msg,
							Oscl::Usb::Ohci::
							BW::Allocator&		allocator
							) throw():
		_context(context),
		_msg(msg),
		_allocator(allocator)
		{
	_context.disablePeriodicList();
	_currentFrameNumber	= _context.getCurrentFrameNumber();
	queueSofTask();
	}

void	Driver::InterruptListDetachDefTask::queueSofTask() throw(){
	DeferredTask<InterruptListDetachDefTask>*
	dt	= new(&_dtMem)
			DeferredTask<	InterruptListDetachDefTask
							>(	*this,
								&Driver::InterruptListDetachDefTask::sofDone
								);
	_context.queueSofTask(*dt);
	}

void	Driver::InterruptListDetachDefTask::sofDone() throw(){
	if(_currentFrameNumber == _context.getCurrentFrameNumber()){
		queueSofTask();
		return;
		}
	_allocator.free(_msg.getPayload()._bwRecToFree);
	_msg.returnToSender();
	_context.enablePeriodicList();
	_context.release(*this);
	}

//================= BulkInListAttachDefTask ========

Driver::BulkInListAttachDefTask::
BulkInListAttachDefTask(	Driver&				context,
							Oscl::Usb::Alloc::
							Pipe::Bulk::IN::
							Req::Api::NewReq&	msg,
							BulkInPipeRec&		pipe
							) throw():
		_context(context),
		_msg(msg),
		_pipe(pipe)
		{
	_context.disableBulkList();
	_currentFrameNumber	= _context.getCurrentFrameNumber();
	queueSofTask();
	}

void	Driver::BulkInListAttachDefTask::queueSofTask() throw(){
	DeferredTask<BulkInListAttachDefTask>*
	dt	= new(&_dtMem)
			DeferredTask<	BulkInListAttachDefTask
							>(	*this,
								&Driver::BulkInListAttachDefTask::sofDone
								);
	_context.queueSofTask(*dt);
	}

void	Driver::BulkInListAttachDefTask::sofDone() throw(){
	if(_currentFrameNumber == _context.getCurrentFrameNumber()){
		queueSofTask();
		return;
		}
	_msg.getPayload().setPipeApi(&_pipe._pipe);
	_msg.returnToSender();
	_context.insertToBulkList(_pipe._pipe.getED());
	_context.enableBulkList();
	_context.release(*this);
	}

//================= BulkInListDetachDefTask ========

Driver::BulkInListDetachDefTask::
BulkInListDetachDefTask(	Driver&				context,
							Oscl::Usb::Alloc::
							Pipe::Bulk::IN::
							Req::Api::FreeReq&	msg
							) throw():
		_context(context),
		_msg(msg)
		{
	_context.disableBulkList();
	_currentFrameNumber	= _context.getCurrentFrameNumber();
	queueSofTask();
	}

void	Driver::BulkInListDetachDefTask::queueSofTask() throw(){
	DeferredTask<BulkInListDetachDefTask>*
	dt	= new(&_dtMem)
			DeferredTask<	BulkInListDetachDefTask
							>(	*this,
								&Driver::BulkInListDetachDefTask::sofDone
								);
	_context.queueSofTask(*dt);
	}

void	Driver::BulkInListDetachDefTask::sofDone() throw(){
	if(_currentFrameNumber == _context.getCurrentFrameNumber()){
		queueSofTask();
		return;
		}
	Oscl::Usb::Pipe::Bulk::IN::PipeApi*
	pipeApi	= &_msg.getPayload().getPipeApi();
	BulkInPipeRec*
	pipe	= _context.removeActiveBulkInPipe(*pipeApi);
	if(pipe){
		_context.removeFromBulkList(pipe->_pipe.getED());
		_context.free(*pipe);
		}
	_msg.returnToSender();
	_context.enableBulkList();
	_context.release(*this);
	}

//================= BulkOutListAttachDefTask ========

Driver::BulkOutListAttachDefTask::
BulkOutListAttachDefTask(	Driver&				context,
							Oscl::Usb::Alloc::
							Pipe::Bulk::OUT::
							Req::Api::NewReq&	msg,
							BulkOutPipeRec&		pipe
							) throw():
		_context(context),
		_msg(msg),
		_pipe(pipe)
		{
	_context.disableBulkList();
	_currentFrameNumber	= _context.getCurrentFrameNumber();
	queueSofTask();
	}

void	Driver::BulkOutListAttachDefTask::queueSofTask() throw(){
	DeferredTask<BulkOutListAttachDefTask>*
	dt	= new(&_dtMem)
			DeferredTask<	BulkOutListAttachDefTask
							>(	*this,
								&Driver::BulkOutListAttachDefTask::sofDone
								);
	_context.queueSofTask(*dt);
	}

void	Driver::BulkOutListAttachDefTask::sofDone() throw(){
	if(_currentFrameNumber == _context.getCurrentFrameNumber()){
		queueSofTask();
		return;
		}
	_msg.getPayload().setPipeApi(&_pipe._pipe);
	_msg.returnToSender();
	_context.insertToBulkList(_pipe._pipe.getED());
	_context.enableBulkList();
	_context.release(*this);
	}

//================= BulkOutListDetachDefTask ========

Driver::BulkOutListDetachDefTask::
BulkOutListDetachDefTask(	Driver&				context,
							Oscl::Usb::Alloc::
							Pipe::Bulk::OUT::
							Req::Api::FreeReq&	msg
							) throw():
		_context(context),
		_msg(msg)
		{
	_context.disableBulkList();
	_currentFrameNumber	= _context.getCurrentFrameNumber();
	queueSofTask();
	}

void	Driver::BulkOutListDetachDefTask::queueSofTask() throw(){
	DeferredTask<BulkOutListDetachDefTask>*
	dt	= new(&_dtMem)
			DeferredTask<	BulkOutListDetachDefTask
							>(	*this,
								&Driver::BulkOutListDetachDefTask::sofDone
								);
	_context.queueSofTask(*dt);
	}

void	Driver::BulkOutListDetachDefTask::sofDone() throw(){
	if(_currentFrameNumber == _context.getCurrentFrameNumber()){
		queueSofTask();
		return;
		}
	Oscl::Usb::Pipe::Bulk::OUT::PipeApi*
	pipeApi	= &_msg.getPayload().getPipeApi();
	BulkOutPipeRec*
	pipe	= _context.removeActiveBulkOutPipe(*pipeApi);
	if(pipe){
		_context.removeFromBulkList(pipe->_pipe.getED());
		_context.free(*pipe);
		}
	_msg.returnToSender();
	_context.enableBulkList();
	_context.release(*this);
	}

