/*
   Copyright (C) 2004 Michael Nelson Moran

   This file is part of OSCL.

   OSCL is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   OSCL is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   You should have received a copy of the GNU General Public License
   along with OSCL; see the file COPYING.  If not, write to the Free
   Software Foundation, 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To negotiate other licensing arrangments/agreements contact
   the copyright holder:

   Michael N. Moran
   5009 Old Field Ct.         mike@mnmoran.org
   Kennesaw, GA, USA 30144    http://mnmoran.org
*/

#ifndef _oscl_frame_stream_rx_pdu_parth_
#define _oscl_frame_stream_rx_pdu_parth_
#include "buffer.h"
#include "fixed.h"
#include "frag.h"
#include "composite.h"
#include "fsm.h"
#include "oscl/mt/itc/srv/closeapi.h"
#include "oscl/mt/itc/srv/close.h"
#include "oscl/pdu/fwd/api.h"
#include "oscl/frame/stream/rx/api.h"

/** */
namespace Oscl {
/** */
namespace Frame {
/** */
namespace Stream {
/** */
namespace RX {
/** */
namespace PDU {

/** This sub-service is responsible for managing the buffer
	resources used to build PDUs from an input stream.
 */
class Part :	private	Oscl::Mt::Itc::Srv::Close::Req::Api,
				public	Oscl::Frame::Stream::RX::Api
				{
	private:
		/** */
		friend class BufferMsg;
		/** */
		friend class FixedMsg;
		/** */
		friend class FragmentMsg;
		/** */
		friend class CompositeMsg;
		/** */
		friend class FreeStoreMgr<Octet>;
		/** */
		friend class FreeStoreMgr<Fixed>;
		/** */
		friend class FreeStoreMgr<Fragment>;
		/** */
		friend class FreeStoreMgr<Composite>;
		/** */
		enum {_minBufferSize=sizeof(BufferMem)};

	private:
		/** */
		Oscl::Mt::Itc::PostMsgApi&	_myPapi;
		/** */
		Oscl::Pdu::FWD::Api&		_fwdApi;
		/** */
		void*						_rawBufferMemory;
		/** */
		unsigned					_bufferSize;
		/** */
		Oscl::Queue<BufferMem>		_bufferPool;
		/** */
		Oscl::Queue<FixedMem>		_fixedPool;
		/** */
		Oscl::Queue<FragmentMem>	_fragmentPool;
		/** */
		Oscl::Queue<CompositeMem>	_compositePool;
		/** */
		FreeStoreMgr<Fixed>			_fixedFSM;
		/** */
		FreeStoreMgr<Fragment>		_fragmentFSM;
		/** */
		FreeStoreMgr<Composite>		_compositeFSM;
		/** */
		FreeStoreMgr<Octet>			_bufferFSM;
		/** */
		const unsigned				_nFragments;
		/** */
		const unsigned				_nComposites;
		/** */
		unsigned					_nFreeBuffer;
		/** */
		unsigned					_nFreeFixed;
		/** */
		unsigned					_nFreeFragment;
		/** */
		unsigned					_nFreeComposite;
		/** */
		Oscl::Mt::Itc::Srv::
		Close::Req::Api::ConcreteSAP		_closeSAP;

		/** */
		CloseReq*							_closeReq;

		/** */
		Oscl::Handle<Oscl::Pdu::Composite>	_currentComposite;
		/** */
		FragmentMem*						_currentFragmentMem;
		/** */
		FixedMem*							_currentFixedMem;
		/** */
		Octet*								_currentBuffer;
		/** */
		unsigned							_currentLength;

	public:
		/** "nFragments" represents the number of elements
			in the fixedMem[] and fragmentMem[] arrays, and
			also represents the number of buffers of size
			"bufferSize" that are contained in "bufferMem".
			
			"bufferMem" constraints:
				o Must be cache aligned
				o Size must be at least sizeof(BufferMem)
				o Size must be even multiple of cache alignment
				o The number of buffers must be the same as
				o The number of bytes in the memory pointed
					to by "bufferMem" must be at least 
					bufferSize*nFragments bytes.
			If any of these constraints are violated, a run-time
			error will happen during the execution of this
			constructor.
		 */
		Part(	Oscl::Mt::Itc::PostMsgApi&			myPapi,
				Oscl::Pdu::FWD::Api&				fwdApi,
				CompositeMem						compositeMem[],
				unsigned							nCompositeMem,
				FragmentMem							fragmentMem[],
				FixedMem							fixedMem[],
				unsigned							nFragments,
				void*								bufferMem,
				unsigned							bufferSize
				) throw();

		/** */
		~Part();

		/** */
		Oscl::Mt::Itc::Srv::Close::Req::Api::SAP&	getCloseSAP() throw();

	private:
		/** */
		void	checkCloseComplete() throw();

		/** */
		bool	allocCurrentFragMem() throw();

		/** */
		void	freeCurrentPduMem() throw();

		/** */
		void	appendCurrentFragment() throw();

	public: // Oscl::Frame::Stream::RX::Api
		/** This operation is invoked when a new frame is being received.
			This operation may be used by the implementation to allocate
			resources for the incomming frame.
		 */
		void		open() throw();

		/** This operation is invoked to append a single octet to
			the open frame. One is returned if the octet is successfuly
			appended. Zero is returned if insufficient resources are
			available to hold the frame, in which case the client
			may elect to abort() the frame.
		 */
		unsigned	append(unsigned char octet) throw();

		/** This operation is invoked to append the specified number
			of octets to the open frame. The number of octets actually
			appended are returned. If the return value is less than
			the number of octets specified by the client, there
			were insufficient resources available to contain the frame,
			and the client should then abort() the frame.
		 */
		unsigned	append(	const void*	octets,
									unsigned	nOctets
									) throw();

		/** This operation is invoked when the last octet in the currently
			open frame has been received and has been determined to be error
			free. The client specifies the number of trailer octets that
			are FCS octets. The implementation may subsequently strip
			the FCS octets from the frame.
		 */
		void		close(unsigned nFcsOctets) throw();

		/** This operation is invoked instead of the close() when the
			currently open frame contains an error.
		 */
		void		abort() throw();

	private:	// allocators
		/** */
		Octet*			allocBuffer() throw();
		/** */
		FixedMem*		allocFixed() throw();
		/** */
		FragmentMem*	allocFragment() throw();
		/** */
		Composite*		allocComposite() throw();

	private:	// deallocation
		/** */
		void	freeBuffer(void *buffer) throw();
		/** */
		void	freeFragment(void* fragment) throw();
		/** */
		void	freeFixed(void* fixed) throw();
		/** */
		void	freeComposite(void* composite) throw();

	private:	// deallocation
		/** */
		void	returnBuffer(BufferMsg* buff) throw();
		/** */
		void	returnFixed(FixedMsg* fixed) throw();
		/** */
		void	returnFragment(FragmentMsg* fragment) throw();
		/** */
		void	returnComposite(CompositeMsg* composite) throw();

	private:	// pdu deallocation actions
		/** */
		void	free(Fragment* fragment) throw();
		/** */
		void	free(Fixed* fixed) throw();
		/** */
		void	free(Composite* composite) throw();
		/** */
		void	free(Octet* buffer) throw();

	private:	// Oscl::Mt::Itc::Srv::Close::Req::Api
		/** */
		void	request(CloseReq& msg) throw();

	private:	// Oscl::Mt::Itc::Srv::Open::Req::Api
		/** */
		void	request(OpenReq& msg) throw();
	};
}
}
}
}
}


#endif
