/*
   Copyright (C) 2004 Michael Nelson Moran

   This file is part of OSCL.

   OSCL is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   OSCL is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   You should have received a copy of the GNU General Public License
   along with OSCL; see the file COPYING.  If not, write to the Free
   Software Foundation, 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To negotiate other licensing arrangments/agreements contact
   the copyright holder:

   Michael N. Moran
   5009 Old Field Ct.         mike@mnmoran.org
   Kennesaw, GA, USA 30144    http://mnmoran.org
*/

#ifndef _oscl_krux_timer_systimerh_
#define _oscl_krux_timer_systimerh_
#include "uktimer.h"
#include "oscl/krux/sema/bqsema.h"
#include "oscl/krux/port/trap.h"
#include "oscl/mt/runnable.h"
#include "oscl/krux/mutex/sema.h"

/** */
namespace Oscl {
/** */
namespace Krux {

/** This class implements the system timer thread which is the
	operating system entity responsible for maintaining timers
	used by the rest of the system.
 */

class SystemTimerThread:	public Mt::Runnable,
							public SystemTimerNode
							{
	private:
		/** The semaphore which is signaled by the clock driver interrupt
			service routine every TOCK.
		 */

		Sema::Base&					_tockSema;

	public:
		/** This is the constructor.
		 */
		SystemTimerThread(Sema::Base& tockSema) throw();

		/** This function implements the Runnable interface and is
			called when the thread begins execution. The main event loop
			of the system timer is contained in this function.
		 */
		void run() throw();

		/** This function is invoked from thread level
		 *  to insert a timer into the SystemTimerThread list.
		 */
		void start(SystemTimer* timer) throw();

		/** This function is invoked from thread level
		 *  to remove a timer from the SystemTimerThread list.
		 */
		void stop(SystemTimer* timer) throw();

		/** This function is invoked from supervisor level
		 *  to insert a timer into the SystemTimerThread list.
		 */
		void suStart(SystemTimer* timer) throw();

		/** This function is invoked from supervisor level
		 *  to remove a timer from the SystemTimerThread list.
		 */
		void suStop(SystemTimer* timer) throw();

		/** This function is invoked from supervisor level
		 *  to process the active timers.
		 */
		void suTock() throw();

		/** This class function is returns the singleton reference
		 *  to the SystemTimerThread.
		 */
		static SystemTimerThread&	GetSystemTimer(void) throw();

	private:
		/** This function is used by the timer thread to insert the
		 *	specified timer into the timer list before the specified timer.
		 */
		void insert(	SystemTimerNode*	nextTimer,
						SystemTimerNode*	newTimer
						) throw();

		/** This function is used by the timer thread to insert the
		 *	specified timer at the end of the timer list.
		 */
		void insert(SystemTimerNode* newTimer) throw();

		/** This function is used by the timer thread to remove the
		 *	specified timer from the timer list
		 */
		void remove(SystemTimerNode* timer) throw();

	public:
		inline SystemTimer*	getTimer() throw() { return 0;}
	};


/** This concrete trap class is used by clients to start
	system timers.
	*/

class StartTimerTrap : public Trap {
	private:
		/** A reference to the system timer thread.
		 */

		SystemTimerThread&	_timerThread;

		/** A pointer to the timer to start.
		 */

		SystemTimer*			_timer;
	public:
		/** The trap constructor.
		 */
		StartTimerTrap(	SystemTimerThread&	timerThread,
						SystemTimer*		timer
						) throw():
			_timerThread(timerThread),_timer(timer){
			}

		/** The trap service routine that is invoked to start the timer
			from the operating system trap.
		 */
		void suService(void) throw();
	};

/** This concrete trap class is used by clients to stop
	system timers.
 */

class StopTimerTrap : public Trap {
	private:
		/** A reference to the system timer thread.
		 */

		SystemTimerThread&	_timerThread;

		/** A pointer to the timer to start.
		 */

		SystemTimer*		_timer;
	public:
		/** The trap constructor.
		 */
		StopTimerTrap(	SystemTimerThread&	timerThread,
						SystemTimer*		timer
						) throw():
			_timerThread(timerThread),_timer(timer){
			}

		/** The trap service routine that is invoked to start the timer
			from the operating system trap.
		 */
		void suService(void) throw();
	};

/** This concrete trap class is used by the system timer thread to
	trap to the operating system and process all pending timers.
 */

class ExpireTimerTrap : public Trap {
	private:
		/** A reference to the timer thread used to envoke
			timer management routines once inside the trap.
		 */
		SystemTimerThread&	_timerThread;
	public:
		/** The timer expirator constructor.
		 */
		ExpireTimerTrap(SystemTimerThread& timerThread) throw():
			_timerThread(timerThread){
			}

		/** The trap service routine that is invoked to begin processing
			timers from the operating system trap.
		 */
		void suService(void) throw();
	};

}
}

#endif
