/*
   Copyright (C) 2004 Michael Nelson Moran

   This file is part of OSCL.

   OSCL is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   OSCL is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   You should have received a copy of the GNU General Public License
   along with OSCL; see the file COPYING.  If not, write to the Free
   Software Foundation, 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To negotiate other licensing arrangments/agreements contact
   the copyright holder:

   Michael N. Moran
   5009 Old Field Ct.         mike@mnmoran.org
   Kennesaw, GA, USA 30144    http://mnmoran.org
*/

#include "integrator.h"

using namespace Oscl;
using namespace Oscl::Logic;

Logic::Lsi::Integ::Integ(	const bool&	dataInput,
							unsigned	nTicks,
							unsigned	upToDownRation
							) throw():
		Logic::Lsi::IntegApi(dataInput),
		_maxValue(nTicks*upToDownRation),
		_currentCount(_maxValue),
		_upCountValue(upToDownRation),
		_downCountValue(1),
		_inhibitInput(*this),
		_setInput(*this),
		_clearInput(*this),
		_inhibited(_inhibitInput.getOutput()),
		_set(_setInput.getOutput()),
		_cleared(_clearInput.getOutput()),
		_integratorState(true)
		{
	_output.asserted();
	}

Seq::InputApi&	Logic::Lsi::Integ::getClockInput() throw(){
	return *this;
	}

Logic::OutputApi&	Logic::Lsi::Integ::getOutput() throw(){
	return _output;
	}

Logic::InputApi&		Logic::Lsi::Integ::getInhibit() throw(){
	return _inhibitInput;
	}

Logic::InputApi&		Logic::Lsi::Integ::getSet() throw(){
	return _setInput;
	}

Logic::InputApi&		Logic::Lsi::Integ::getClear() throw(){
	return _clearInput;
	}

void	Logic::Lsi::Integ::attach(	Seq::SubjectApi&	output,
									Seq::Wire&		wire
									) throw(){
	wire.setSeqApi(*this);
	output.attach(wire);
	}

void	Logic::Lsi::Integ::detach(	Seq::SubjectApi&	output,
									Seq::Wire&		wire
									) throw(){
	output.detach(wire);
	}

void	Logic::Lsi::Integ::tick() throw(){
	if(_inhibited || _set || _cleared){
		return;
		}
	if(_dataInput){
		if(_currentCount >= _maxValue) return;
		_currentCount	+= _upCountValue;
		if(_currentCount >= _maxValue){
			_currentCount		= _maxValue;
			_integratorState	= true;
			}
		}
	else{
		if(_currentCount <= _minValue) return;
		_currentCount	-= _downCountValue;
		if(_currentCount <= _minValue){
			_currentCount	= _minValue;
			_integratorState	= false;
			}
		}
	evaluateOutput();
	}

void	Logic::Lsi::Integ::clearAsserted() throw(){
	_currentCount		= _minValue;
	_integratorState	= false;
	evaluateOutput();
	}

void	Logic::Lsi::Integ::clearNegated() throw(){
	if(_set)	setAsserted();
	else		evaluateOutput();
	}

void	Logic::Lsi::Integ::setAsserted() throw(){
	if(_cleared) return;
	_currentCount		= _maxValue;
	_integratorState	= true;
	evaluateOutput();
	}

void	Logic::Lsi::Integ::setNegated() throw(){
	evaluateOutput();
	}

void	Logic::Lsi::Integ::evaluateOutput() throw() {
	if(_cleared){
		_output.negated();
		return;
		}
	if(_set){
		_output.asserted();
		return;
		}
	if(_integratorState){
		_output.asserted();
		}
	else{
		_output.negated();
		}
	}

////////// Inhibit

Logic::Lsi::Integ::Inhibit::Inhibit(Logic::Lsi::Integ& context) throw():
		_context(context),
		_output(false),
		_assertedCount(0)
		{
	}

const bool&	Logic::Lsi::Integ::Inhibit::getOutput() throw(){
	return _output;
	}

void	Logic::Lsi::Integ::Inhibit::attach(	Logic::OutputApi&	output,
											Logic::Wire&		wire
											) throw(){
	wire.setStateApi(*this);
	if(!output.getOutputRef()){
		++_assertedCount;
		}
	output.attach(wire);
	}

void	Logic::Lsi::Integ::Inhibit::detach(	Logic::OutputApi&	output,
											Logic::Wire&		wire
											) throw(){
	output.detach(wire);
	if(output.getOutputRef()) negated();
	}

void	Logic::Lsi::Integ::Inhibit::asserted() throw(){
	++_assertedCount;
	_output	= true;
	}

void	Logic::Lsi::Integ::Inhibit::negated() throw(){
	--_assertedCount;
	if(!_assertedCount) _output = false;
	}

////////// Clear

Logic::Lsi::Integ::Clear::Clear(Logic::Lsi::Integ& context) throw():
		_context(context),
		_output(false),
		_assertedCount(0)
		{
	}

const bool&	Logic::Lsi::Integ::Clear::getOutput() throw(){
	return _output;
	}

void	Logic::Lsi::Integ::Clear::attach(	Logic::OutputApi&	output,
											Logic::Wire&			wire
											) throw(){
	wire.setStateApi(*this);
	if(!output.getOutputRef()){
		++_assertedCount;
		}
	output.attach(wire);
	}

void	Logic::Lsi::Integ::Clear::detach(	Logic::OutputApi&	output,
											Logic::Wire&			wire
											) throw(){
	output.detach(wire);
	if(output.getOutputRef()) negated();
	}

void	Logic::Lsi::Integ::Clear::asserted() throw(){
	++_assertedCount;
	_output	= true;
	_context.clearAsserted();
	}

void	Logic::Lsi::Integ::Clear::negated() throw(){
	--_assertedCount;
	if(_assertedCount) return;
	_output	= false;
	_context.clearNegated();
	}

////////// Set

Logic::Lsi::Integ::Set::Set(Logic::Lsi::Integ& context) throw():
		_context(context),
		_output(false),
		_assertedCount(0)
		{
	}

const bool&	Logic::Lsi::Integ::Set::getOutput() throw(){
	return _output;
	}

void	Logic::Lsi::Integ::Set::attach(	Logic::OutputApi&	output,
										Logic::Wire&			wire
										) throw(){
	wire.setStateApi(*this);
	if(!output.getOutputRef()){
		++_assertedCount;
		}
	output.attach(wire);
	}

void	Logic::Lsi::Integ::Set::detach(	Logic::OutputApi&	output,
										Logic::Wire&			wire
										) throw(){
	output.detach(wire);
	if(output.getOutputRef()) negated();
	}

void	Logic::Lsi::Integ::Set::asserted() throw(){
	++_assertedCount;
	_output	= true;
	_context.setAsserted();
	}

void	Logic::Lsi::Integ::Set::negated() throw(){
	--_assertedCount;
	if(_assertedCount) return;
	_output	= false;
	_context.setNegated();
	}

