/*
   Copyright (C) 2004 Michael Nelson Moran

   This file is part of OSCL.

   OSCL is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   OSCL is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   You should have received a copy of the GNU General Public License
   along with OSCL; see the file COPYING.  If not, write to the Free
   Software Foundation, 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To negotiate other licensing arrangments/agreements contact
   the copyright holder:

   Michael N. Moran
   5009 Old Field Ct.         mike@mnmoran.org
   Kennesaw, GA, USA 30144    http://mnmoran.org
*/

#include <stdlib.h>
#include "thread.h"
#include "oscl/timer/busywait.h"
#include "oscl/timer/tocks.h"
#include "oscl/error/fatal.h"

using namespace Oscl;

static void entryPoint(cyg_addrword_t runnable);

static bool            threadingEnabled;	// false
static bool            keyCreated=false;
static cyg_ucount32    key;

eCosThread::eCosThread(    Mt::Runnable&   runnable,
                           const char*     name,
                           cyg_ucount32    priority,
                           cyg_ucount32    stackSize
                            ) throw():
        _runnable(runnable),
        _syncSema(),
        _name(name)
        {

    stackSize = stackSize == 0 ? 1 : stackSize;

    _stack = (char *) malloc (stackSize);

    if(!_stack)
        ErrorFatal::logAndExit(
                "Mt::eCosThread::eCosThread() out of memory.\n");

    cyg_thread_create (priority,              // priority
                       entryPoint,            // entry point
                       (cyg_addrword_t) this, // argument
                       (char *) name,         // overriding constant not good
                       (char*)_stack,         // stack base 
                       stackSize,             // task stack size
                       &_handle,              // handle to task
                       &_thread);             // thread object

    cyg_thread_resume (_handle);

    }

eCosThread::eCosThread(	Mt::Runnable&	runnable,
						void*			stack,
						cyg_ucount32	stackSize,
						cyg_ucount32	priority  = DEFAULT_TASK_PRIORITY,
						const char*		name=""
						) throw():
        _runnable(runnable),
        _syncSema(),
        _name(name),
		_stack(stack)
        {
    cyg_thread_create (priority,              // priority
                       entryPoint,            // entry point
                       (cyg_addrword_t) this, // argument
                       (char *) name,         // overriding constant not good
                       _stack,                // stack base 
                       stackSize,             // task stack size
                       &_handle,              // handle to task
                       &_thread);             // thread object

    cyg_thread_resume (_handle);
    }

eCosThread::~eCosThread(){
    cyg_thread_kill(_handle);
    free(_stack);
    }

void    eCosThread::suspend() throw(){
    ErrorFatal::logAndExit("eCosThread::suspend() not implemented.\n");
    }

void    eCosThread::suSuspend() throw(){
    ErrorFatal::logAndExit("eCosThread::suSuspend() not implemented.\n");
    }

void    eCosThread::resume() throw(){
    ErrorFatal::logAndExit("eCosThread::resume() not implemented.\n");
    }

void    eCosThread::suResume() throw(){
    ErrorFatal::logAndExit("eCosThread::suResume() not implemented.\n");
    }

Mt::SignalApi&    eCosThread::getSyncSignalApi() throw(){
    return _syncSema;
    }

void        eCosThread::syncWait() throw(){
    _syncSema.wait();
    }

const char* eCosThread::getName() throw(){
    return _name.getString();
    }

//==============================================================================

static void entryPoint(cyg_addrword_t vthread){
    eCosThread*        thread    = (eCosThread*)vthread;
    Mt::Thread*            thisThread    = thread;

    cyg_thread_set_data(key, (CYG_ADDRWORD) thisThread);
	threadingEnabled	= true;

    thread->_runnable.run();
    }

Mt::Thread::Thread() throw(){
    if(!keyCreated){

        key = cyg_thread_new_data_index();

        keyCreated=true;
        }
    }

Mt::Thread::~Thread(){
    }

Mt::Thread&    Mt::Thread::getCurrent() throw() {
    Mt::Thread*    thread             = (Mt::Thread*)cyg_thread_get_data(key);
    return *thread;
    }

void    Mt::Thread::sleep(unsigned long milliseconds) throw(){
	// Presumably, keyCreated means the kernel has been initialized
	// and we are multi-threading.
	if(threadingEnabled){
		cyg_thread_delay(OsclTimerMillisecondsToTocks(milliseconds));
		}
	else{
		OsclTimerBusyWaitDelay(milliseconds);
		}
    }

