/*
   Copyright (C) 2004 Michael Nelson Moran

   This file is part of OSCL.

   OSCL is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   OSCL is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   You should have received a copy of the GNU General Public License
   along with OSCL; see the file COPYING.  If not, write to the Free
   Software Foundation, 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To negotiate other licensing arrangments/agreements contact
   the copyright holder:

   Michael N. Moran
   5009 Old Field Ct.         mike@mnmoran.org
   Kennesaw, GA, USA 30144    http://mnmoran.org
*/

#ifndef _oscl_mt_ecos_threadh_
#define _oscl_mt_ecos_threadh_
#include <cyg/kernel/kapi.h>
#include "oscl/mt/runnable.h"
#include "oscl/mt/thread.h"
#include "sema.h"
#include "oscl/strings/fixed.h"

#define DEFAULT_TASK_STACK_SIZE 2048
#define DEFAULT_TASK_PRIORITY 5

/** */
namespace Oscl {

/** This concrete class implements a task,
	which provides a concurrency context for a
	Mt::Runnable object, in terms of a Posix pthread.
	NOTE: This task must only be dynamically allocated
	since it performs dynamic resource allocation
	and requires kernel services to be operating.
 */

class eCosThread : public Mt::Thread {
	public:
		/** Runnable interface.
		 */
		Mt::Runnable&			_runnable;

	private:
		/** The thread identifier.
		 */
		cyg_thread				_thread;

		/** The thread handle.
		*/
		cyg_handle_t			_handle;

		/** The thread synchronized message semaphore.
		 */
		eCosSemaphore			_syncSema;

		/** Thread name.
		 */
		Strings::Fixed<32>			_name;

		/** define our stack 
		*/
		void*					_stack;

	public: //Constructors
		/** The constructor requires: a runnable object which is started
			before the constructor completes, a short string describing this
			task instance; a priority ranging from high priority (0) to low
			priority (10); and a stack size that indicates the size of the
			stack in bytes allocated for this task.
		 */
		eCosThread(	Mt::Runnable&	runnable,
					const char*		name,
					cyg_ucount32	priority  = DEFAULT_TASK_PRIORITY,
					cyg_ucount32	stackSize = DEFAULT_TASK_STACK_SIZE
					) throw();
		/** */
		eCosThread(	Mt::Runnable&	runnable,
					void*			stack,
					cyg_ucount32	stackSize,
					cyg_ucount32	priority  = DEFAULT_TASK_PRIORITY,
					const char*		name=""
					) throw();

		/** Virtual destructors rock!
		 */
		virtual ~eCosThread();

		/** Suspend this thread from the task level.
		 */
		void	suspend() throw();

		/** Suspend this thread from the interrupt level.
		 */
		void	suSuspend() throw();

		/** Resume this thread from the task level.
		 */
		void	resume() throw();

		/** Resume this thread from the interrupt level.
		 */
		void	suResume() throw();

		/** This operation returns the signal interface for the thread
			semaphore, which is used for synchronous message transactions.
			This is the sister operation to the thread's syncWait()
			operation.
		 */
		Mt::Sema::SignalApi&	getSyncSignalApi() throw();

		/** This operation blocks on the thread semaphore, waiting for
			a synchronous message operation to complete. This is the
			sister operation to the thread's getSyncSignalApi() operation.
		 */
		void		syncWait() throw();

		/** This operation returns the name (null terminated string)
			of the current thread.
		 */
		virtual const char*	getName() throw();
	};

};

#endif
