/*
   Copyright (C) 2004 Michael Nelson Moran

   This file is part of OSCL.

   OSCL is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   OSCL is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   You should have received a copy of the GNU General Public License
   along with OSCL; see the file COPYING.  If not, write to the Free
   Software Foundation, 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To negotiate other licensing arrangments/agreements contact
   the copyright holder:

   Michael N. Moran
   5009 Old Field Ct.         mike@mnmoran.org
   Kennesaw, GA, USA 30144    http://mnmoran.org
*/

#ifndef _oscl_mt_itc_dyn_core_coreh_
#define _oscl_mt_itc_dyn_core_coreh_
#include "creator.h"
#include "oscl/mt/itc/dyn/fsm/fsm.h"
#include "oscl/error/fatal.h"
#include "cli.h"
#include "ref.h"

/** */
namespace Oscl {
/** */
namespace Mt {
/** */
namespace Itc {
/** */
namespace Dyn {

/** This abstract template implements the core behavior of
	a Dynamic Service. Concrete Dynamic Service sub classes
	must implement openDynamicService(), closeDynamicService(),
	and getDynSrv().
 */
template <class dynSrvType>
class Core :
		public Creator::Req::Api<dynSrvType>,
		public StateVar::ContextApi
		{
	private:
		/** */
		Dyn::StateVar						_state;
		/** */
		Queue<typename Dyn::Cli::Req::Api<dynSrvType>::ReleaseReq>	_notificationList;
		/** */
		unsigned long						_refCount;
		/** */
		Mt::Itc::Srv::Open::Req::Api::
		OpenReq*							_currentOpenReq;
		/** */
		typename Dyn::Ref::Req::Api<dynSrvType>::
		IncrReq*							_currentIncrReq;
		/** */
		typename Dyn::Ref::Req::Api<dynSrvType>::
		DecrReq*							_currentDecrReq;
		/** */
		typename Dyn::Cli::Req::Api<dynSrvType>::
		CancelReleaseReq*					_currentCancelReq;
		/** */
		typename Dyn::Creator::Req::Api<dynSrvType>::
		DeactivateReq*						_currentDeactivateReq;
		/** */
		Mt::Itc::Srv::Close::Req::Api::
		CloseReq*							_currentCloseReq;
	public:
		/** */
		Core(Mt::Itc::PostMsgApi& dynSrv) throw();
	public:	// Dyn::StateVar::ContextApi
		/** Invoked by the Dynamic Service FSM.
			Invokes the openDynamicService() operation.
		 */
		void	activateDynamicService() throw();
		/** Invoked by the Dynamic Service FSM.
			Invokes the closeDynamicService() operation.
		 */
		void	deactivateDynamicService() throw();
		/** Invoked by the Dynamic Service FSM.
			Invokes the returnToSender() of the current
			open request.
		 */
		void	finishedWithOpenReq() throw();
		/** Invoked by the Dynamic Service FSM.
			Invokes the returnToSender() of the current
			prepare for close request.
		 */
		void	finishedWithClosePrepReq() throw();
		/** Invoked by the Dynamic Service FSM.
			Invokes the returnToSender() of the current
			close request.
		 */
		void	finishedWithCloseReq() throw();
		/** Invoked by the Dynamic Service FSM.
			Searches the notification list for the notification
			request specified by the current notification
			cancelation request; removes it from the list; and
			then invokes its returnToSender() operation.
		 */
		void	cancelNotificationReq() throw();
		/** Invoked by the Dynamic Service FSM.
			Invokes the returnToSender() operation of the
			current reference count increment request.
		 */
		void	returnIncrReq() throw();
		/** Invoked by the Dynamic Service FSM.
			Invokes the returnToSender() operation of the
			current reference count decrement request.
		 */
		void	returnDecrReq() throw();
		/** Invoked by the Dynamic Service FSM.
			Increments the reference count.
		 */
		void	incrementRefCount() throw();
		/** Invoked by the Dynamic Service FSM.
			Decrements the reference count.
		 */
		void	decrementRefCount() throw();
		/** Invoked by the Dynamic Service FSM.
			Removes each notification request from the
			notification list and invokes the returnToSender()
			on each.
		 */
		void	notifyClosing() throw();
		/** Invoked by the Dynamic Service FSM.
			Returns true if the reference count is zero.
		 */
		bool	refCountIsZero() throw();
	private:
		/** Invoked when an open request is received.
		 */
		void	request(	Mt::Itc::Srv::Open::Req::Api::
							OpenReq&			msg
							) throw();
		/** Invoked when a close request is received.
		 */
		void	request(	Mt::Itc::Srv::Close::Req::Api::
							CloseReq&			msg
							) throw();
		/** Invoked when a request to increment the reference
			count is received.
		 */
		void	request(	typename Dyn::Ref::Req::Api<dynSrvType>::
							IncrReq&			msg
							) throw();
		/** Invoked when a request to decrement the reference
			count is received.
		 */
		void	request(	typename Dyn::Ref::Req::Api<dynSrvType>::
							DecrReq&			msg
							) throw();
		/** Invoked when a release notification request is received.
		 */
		void	request(	typename Dyn::Cli::Req::Api<dynSrvType>::
							ReleaseReq&			msg
							) throw();
		/** Invoked when a release notification cancelation
			request is received.
		 */
		void	request(	typename Dyn::Cli::Req::Api<dynSrvType>::
							CancelReleaseReq&	msg
							) throw();
		/** Invoked when a prepare to close request is received.
		 */
		void	request(	typename Dyn::Creator::Req::Api<dynSrvType>::
							DeactivateReq&		msg
							) throw();
	protected:
		/** This operation is implemented by sub-classes.
			The implementation takes whatever action is necessary
			to "open" the sub-class. This typically entails binding
			to lower layer services. Function call semantics are
			assumed, and thus ITC must be done synchronously.
		 */
		virtual void	openDynamicService() throw()=0;
		/** This operation is implemented by sub-classes.
			The implementation takes whatever action it desires
			to prepare the sub-class for a subsequent
			closeDynamicService() operation. Function call
			semantics are assumed, and thus ITC must be done
			synchronously.
		 */
//		virtual void	prepareToCloseDynamicService() throw()=0;
		/** This operation is implemented by sub-classes.
			The implementation takes whatever action is necessary
			to "close" the sub-class in preparation for subsequent
			destruction. Function call semantics are assumed,
			and thus ITC must be done synchronously.
		 */
		virtual void	closeDynamicService() throw()=0;
		/** This operation is implemented by sub-classses.
			It returns a reference to the actual Dynamic Service.
		 */
		virtual dynSrvType&		getDynSrv() throw()=0;
	};

template <class dynSrvType>
Core<dynSrvType>::Core(Mt::Itc::PostMsgApi& papi) throw():
		Creator::Req::Api<dynSrvType>(papi),
		_state(*this),
		_refCount(0)
		{
	}

template <class dynSrvType>
void	Core<dynSrvType>::activateDynamicService() throw(){
	openDynamicService();
	}

//template <class dynSrvType>
//void	Core<dynSrvType>::prepareForDeactivation() throw(){
//	prepareToCloseDynamicService();
//	}

template <class dynSrvType>
void	Core<dynSrvType>::deactivateDynamicService() throw(){
	closeDynamicService();
	}

template <class dynSrvType>
void	Core<dynSrvType>::finishedWithOpenReq() throw(){
	_currentOpenReq->returnToSender();
	}

template <class dynSrvType>
void	Core<dynSrvType>::finishedWithClosePrepReq() throw(){
	_currentDeactivateReq->returnToSender();
	}

template <class dynSrvType>
void	Core<dynSrvType>::finishedWithCloseReq() throw(){
	_currentCloseReq->returnToSender();
	}

template <class dynSrvType>
void	Core<dynSrvType>::cancelNotificationReq() throw(){
	_notificationList.remove(&_currentCancelReq->getPayload().getReqToCancel());
	_currentCancelReq->getPayload().getReqToCancel().returnToSender();
	_currentCancelReq->returnToSender();
	}

template <class dynSrvType>
void	Core<dynSrvType>::returnIncrReq() throw(){
	_currentIncrReq->returnToSender();
	}

template <class dynSrvType>
void	Core<dynSrvType>::returnDecrReq() throw(){
	_currentDecrReq->returnToSender();
	}

template <class dynSrvType>
void	Core<dynSrvType>::incrementRefCount() throw(){
	++_refCount;
	}

template <class dynSrvType>
void	Core<dynSrvType>::decrementRefCount() throw(){
	--_refCount;
	}

template <class dynSrvType>
void	Core<dynSrvType>::notifyClosing() throw(){
	typename Dyn::Cli::Req::Api<dynSrvType>::ReleaseReq* next;
	while((next=_notificationList.get())){
		next->returnToSender();
		}
	}

template <class dynSrvType>
bool	Core<dynSrvType>::refCountIsZero() throw(){
	return !_refCount;
	}

template <class dynSrvType>
void	Core<dynSrvType>::request(	Mt::Itc::Srv::Open::
									Req::Api::OpenReq&		msg
									) throw(){
	_currentOpenReq	= &msg;
	_state.open();
	}

template <class dynSrvType>
void	Core<dynSrvType>::request(	Mt::Itc::Srv::Close::
									Req::Api::CloseReq&		msg
									) throw(){
	_currentCloseReq	= &msg;
	_state.close();
	}

template <class dynSrvType>
void	Core<dynSrvType>::request(	typename Dyn::Ref::Req::
									Api<dynSrvType>::IncrReq&	msg
									) throw(){
	_currentIncrReq	= &msg;
	_state.incrementRefCount();
	}

template <class dynSrvType>
void	Core<dynSrvType>::request(	typename Dyn::Ref::Req::
									Api<dynSrvType>::DecrReq&	msg
									) throw(){
	_currentDecrReq	= &msg;
	_state.decrementRefCount();
	}

template <class dynSrvType>
void	Core<dynSrvType>::request(	typename Dyn::Cli::Req::
									Api<dynSrvType>::ReleaseReq&	msg
									) throw(){
	_notificationList.put(&msg);
	_state.notificationReq();
	}

template <class dynSrvType>
void	Core<dynSrvType>::request(	typename Dyn::Cli::Req::
									Api<dynSrvType>::CancelReleaseReq&	msg
									) throw(){
	_currentCancelReq	= &msg;
	_state.cancelNotificationReq();
	}

template <class dynSrvType>
void	Core<dynSrvType>::request(	typename Dyn::Creator::Req::
									Api<dynSrvType>::DeactivateReq&	msg
									) throw(){
	_currentDeactivateReq	= &msg;
	_state.deactivateReq();
	}

}
}
}
}

#endif
