/*
   Copyright (C) 2004 Michael Nelson Moran

   This file is part of OSCL.

   OSCL is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   OSCL is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   You should have received a copy of the GNU General Public License
   along with OSCL; see the file COPYING.  If not, write to the Free
   Software Foundation, 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To negotiate other licensing arrangments/agreements contact
   the copyright holder:

   Michael N. Moran
   5009 Old Field Ct.         mike@mnmoran.org
   Kennesaw, GA, USA 30144    http://mnmoran.org
*/

#include <string.h>
#include "adaptor.h"
#include "oscl/endian/type.h"
#include "oscl/bignum/value.h"
#include "oscl/des/cbc.h"
#include "oscl/entropy/random.h"

using namespace Oscl::SS;

static const unsigned char	diffieHellmanPrime[]	= {
	0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
	0xC9,0x0F,0xDA,0xA2,0x21,0x68,0xC2,0x34,
	0xC4,0xC6,0x62,0x8B,0x80,0xDC,0x1C,0xD1,
	0x29,0x02,0x4E,0x08,0x8A,0x67,0xCC,0x74,
	0x02,0x0B,0xBE,0xA6,0x3B,0x13,0x9B,0x22,
	0x51,0x4A,0x08,0x79,0x8E,0x34,0x04,0xDD,
	0xEF,0x95,0x19,0xB3,0xCD,0x3A,0x43,0x1B,
	0x30,0x2B,0x0A,0x6D,0xF2,0x5F,0x14,0x37,
	0x4F,0xE1,0x35,0x6D,0x6D,0x51,0xC2,0x45,
	0xE4,0x85,0xB5,0x76,0x62,0x5E,0x7E,0xC6,
	0xF4,0x4C,0x42,0xE9,0xA6,0x37,0xED,0x6B,
	0x0B,0xFF,0x5C,0xB6,0xF4,0x06,0xB7,0xED,
	0xEE,0x38,0x6B,0xFB,0x5A,0x89,0x9F,0xA5,
	0xAE,0x9F,0x24,0x11,0x7C,0x4B,0x1F,0xE6,
	0x49,0x28,0x66,0x51,0xEC,0xE6,0x53,0x81,
	0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF
	};
static const unsigned char	diffieHellmanG[]	= {
	0x02
	};

enum{hugeSize=2*bigNumSize};
enum{smallSize=bigNumSize/2};

static const Oscl::BigNum::Value<bigNumSize>
	diffieHellmanPrimeConst(	diffieHellmanPrime,
								sizeof(diffieHellmanPrime)
								);

static const Oscl::BigNum::Value<bigNumSize>
	diffieHellmanGConst(	diffieHellmanG,
							sizeof(diffieHellmanG)
							);

Adaptor::Adaptor(	Oscl::Stream::Input::Api&	streamIn,
					Oscl::Stream::Output::Api&	streamOut
					) throw():
		_streamIn(streamIn),
		_streamOut(streamOut),
		_open(false),
		_txLength(2),
		_rxLength(0),
		_rxOffset(0),
		_remainingInPacket(0)
		{
	}

void	Adaptor::open() throw(){
	_secretRandomNumber	= OsclEntropyGetRandom();
	Oscl::BigNum::Value<bigNumSize> privateKey;
	privateKey	= _secretRandomNumber;
	Oscl::BigNum::Value<bigNumSize> publicKey;
	publicKey	= diffieHellmanGConst;
	publicKey.powXmodY(privateKey,diffieHellmanPrimeConst);

	unsigned	len;
	len	= _streamOut.write(&publicKey,sizeof(publicKey));
	if(len < sizeof(publicKey)){
		_open	= false;
		return;
		}

	_streamOut.flush();

	unsigned		remaining	= keySize;
	len	= 0;
	while(remaining){
		len=_streamIn.read(&_peerPublicKey[keySize-remaining],remaining);
		if(len == 0){
			_open	= false;
			return;
			}
		remaining	-= len;
		}

	_sessionKey	= Oscl::BigNum::Value<bigNumSize>(_peerPublicKey,keySize);

	_sessionKey.powXmodY(privateKey,diffieHellmanPrimeConst);
	_open	= true;
	}

void	Adaptor::close() throw(){
	_open	= false;
	}

unsigned long	Adaptor::read(	void*			dest,
								unsigned long	maxSize
								) throw(){
	while(true){
		if(!_open) return 0;
		unsigned	len	= copyOut(dest,maxSize);
		if(len){
			return len;
			}
		recv();
		}
	return 0;
	}

unsigned long	Adaptor::read(Buffer::Base& dest) throw(){
	if(!_open) return 0;
	return 0;
	}

//
// Packet Format:
// * All packets must be an even multiple of 8 bytes including
//   the length field.
// * Packets must be at least 8 bytes long.
// * Any padding is always in the last bytes of the packet.
// * The length field is 32 bits Big Endian.
// * The length field specifies the length of the encrypted
//   data not including the length field or the padding.
// * Padding length is implicit since.
// * The maximum packet size is (2**16)-1 bytes.
// * The maximum payload size is thus 65,526 bytes.
// * The maximum value of the length fields is thus 65,526.
// * The minimum packet size is 8 bytes.
// * The minimum payload size is 1 bytes.
// * The minimum value of the length fields is 1.

unsigned long	Adaptor::write(	const void*			source,
								unsigned long		count
								) throw(){
	if(!_open){
		return 0;
		}
	const unsigned char*	p	= (const unsigned char*)source;
	for(unsigned long i=0;count;){
		unsigned	len	= append(&p[i],count);
		if(len < count){
			send();
			if(!_open){
				return i;
				}
			}
		i		+= len;
		count	-= len;
		}
	return 0;
	}

unsigned long	Adaptor::write(const Buffer::Base& source) throw(){
	return write(source.startOfData(),source.length());
	}

void	Adaptor::flush() throw(){
	if(!_open) return;
	if(_txLength > 2){
		send();
		}
	_streamOut.flush();
	}

unsigned	Adaptor::append(	const void*	source,
								unsigned	length
								) throw(){
	unsigned	remaining	= txBufferSize - _txLength;
	if(length > remaining){
		length	= remaining;
		}
	memcpy(&_txBuffer[_txLength],source,length);
	_txLength	+= length;
	return length;
	}

void	Adaptor::send() throw(){
	Oscl::Endian::Big::U16*	length	= (Oscl::Endian::Big::U16*)_txBuffer;
	*length	= (_txLength - 2);
	unsigned	nBlocks	= ((_txLength)+7)/8;
	unsigned	size	= (nBlocks)*8;
	unsigned	len;
	len	= Oscl::DES::CBC::encrypt(	(const unsigned char*)&_sessionKey,
									_txBuffer,
									size,
									_txBuffer,
									size
									);
	if(len < size){
		// This should NEVER happen
		for(;;);
		}

	len	= _streamOut.write(_txBuffer,size);
	if(len < size){
		_open	= false;
		}
	_txLength		= sizeof(Oscl::Endian::Big::U16);
	}

void	Adaptor::recv() throw(){
	if(_rxLength) return;

	unsigned long	offset	= 0;
	if(!_remainingInPacket){
		while(!_remainingInPacket){
			unsigned	len;
			// Read length field (minimum packet)
			for(unsigned i= 0;i<8;){
				len	= _streamIn.read(&_rxBuffer[i],(8-i));
				if(!len){
					_rxLength	= 0;
					_open		= false;
					return;
					}
				i	+= len;
				}
			len	= Oscl::DES::CBC::decrypt(	(const unsigned char*)&_sessionKey,
											_rxBuffer,
											8,
											_rxBuffer,
											8
											);
			_rxOffset	= 2;
			_rxLength	= 6;
			Oscl::Endian::Big::U16*	length	= (Oscl::Endian::Big::U16*)_rxBuffer;
			unsigned	pLength	= *length;
			unsigned	nBlocks	= ((pLength+2)+7)/8;
			_remainingInPacket	= nBlocks*8;
			_rxPadding			= _remainingInPacket-(pLength+2);
			_remainingInPacket	-= 8;
			}
		offset	= 8;
		}
	else {
		_rxLength	= 0;
		_rxOffset	= 0;
		}

	unsigned	nToRead		= _remainingInPacket;
	unsigned	remaining	= rxBufferSize-(_rxOffset+_rxLength);
	if(nToRead > remaining){
		nToRead	=	remaining;
		}

	for(unsigned i=0;i<nToRead;){
		unsigned	nRead	= _streamIn.read(&_rxBuffer[_rxOffset+_rxLength],nToRead);
		if(!nRead){
			_open	= false;
			return;
			}
		_rxLength	+= nRead;
		i			+= nRead;
		}
	_remainingInPacket	-= nToRead;

	unsigned len;
	len	= Oscl::DES::CBC::decrypt(	(const unsigned char*)&_sessionKey,
									&_rxBuffer[offset],
									nToRead,
									&_rxBuffer[offset],
									nToRead
									);
	if(!_remainingInPacket){
		_rxLength	-= _rxPadding;
		}
	}

unsigned	Adaptor::copyOut(	void*		b,
								unsigned	maxBufferSize
								) throw(){
	if(maxBufferSize < _rxLength){
		maxBufferSize	= _rxLength;
		}
	else{
		maxBufferSize	= _rxLength;
		}
	if(!maxBufferSize) return 0;
	unsigned char*	buffer	= (unsigned char*)b;
	memcpy(buffer,&_rxBuffer[_rxOffset],maxBufferSize);
	_rxOffset	+= maxBufferSize;
	_rxLength	-= maxBufferSize;
	return maxBufferSize;
	}

