/*
   Copyright (C) 2004 Michael Nelson Moran

   This file is part of OSCL.

   OSCL is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   OSCL is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   You should have received a copy of the GNU General Public License
   along with OSCL; see the file COPYING.  If not, write to the Free
   Software Foundation, 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To negotiate other licensing arrangments/agreements contact
   the copyright holder:

   Michael N. Moran
   5009 Old Field Ct.         mike@mnmoran.org
   Kennesaw, GA, USA 30144    http://mnmoran.org
*/

#include "api.h"
#include <string.h>
#include <stdio.h>
#ifndef __GNUC__
#ifdef __ENEA_OSE__
#include <outfmt.h>
#else
#error	"need definition of snprintf()"
#endif
#endif

using namespace Oscl::Strings;

/**
	To support both dynamic and static strings, I will require dynamic
	strings to polymorphically override the appropriate members and
	then use the base operators subsequent to buffer allocation and
	expansion. Thus, this base class will implement the static
	functionality which is common to both static and dynmic string
	manipulation.
	We can assume and require that every modification operation
	results in a null terminated string.
 */

inline void	Api::append(const char* src) throw(){
	char*			p			= getBuffer();
	unsigned long	maxLen		= maxLength();
	unsigned long	curLen		= length();
	unsigned long	remaining	= maxLen - curLen;
	p[maxLen]	= '\0';	// This is cleared here to enable fast truncation check.
	strncat(p,src,remaining);
	if(p[maxLen]){
		p[maxLen]	= '\0';
		_truncated	= true;
		}
	else{
		_truncated	= false;
		}
	}

inline void	Api::assign(const char* src) throw(){
	char*			p			= getBuffer();
	unsigned long	maxLen		= maxLength();
	p[maxLen]	= '\0';	// This is cleared here to enable fast truncation check.
	strncpy(p,src,maxLen);
	if(p[maxLen]){
		p[maxLen]	= '\0';
		_truncated	= true;
		}
	else{
		_truncated	= false;
		}
	}

Api::Api() throw():
		_truncated(false)
		{
	}

Api&	Api::operator += (signed long num) throw(){
	// big enough for 128-bit signed long , sign, and null
	static const size_t	decimalBuffSize	= 39+1+1;
	char			b[decimalBuffSize];
	sprintf(b,"%ld",num);
	append(b);
	return *this;
	}

Api&	Api::operator += (signed num) throw(){
	// big enough for 128-bit signed long , sign, and null
	static const size_t	decimalBuffSize	= 39+1+1;
	char			b[decimalBuffSize];
	snprintf(b,decimalBuffSize,"%d",num);
	append(b);
	return *this;
	}

Api&	Api::operator += (unsigned long num) throw(){
	// big enough for 128-bit unsigned long and null
	static const size_t	decimalBuffSize	= 39+1;
	char			b[decimalBuffSize];
	snprintf(b,decimalBuffSize,"%lu",num);
	append(b);
	return *this;
	}

Api&	Api::operator += (unsigned num) throw(){
	// big enough for 128-bit unsigned and null
	static const size_t	decimalBuffSize	= 39+1;
	char			b[decimalBuffSize];
	snprintf(b,decimalBuffSize,"%u",num);
	append(b);
	return *this;
	}

Api&	Api::operator += (const Api& s) throw(){
	append(s.getString());
	return *this;
	}

Api&	Api::operator += (const char* s) throw(){
	append(s);
	return *this;
	}

Api&	Api::operator = (const char* s) throw(){
	assign(s);
	return *this;
	}

Api&	Api::operator = (const Api& s) throw(){
	assign(s.getString());
	return *this;
	}

bool	Api::operator == (const char* s) const throw(){
	return (strcmp(s,getString())==0);
	}

bool	Api::operator == (const Api& s) const throw(){
	return (strcmp(s.getString(),getString())==0);
	}

bool	Api::operator != (const char* s) const throw(){
	return (strcmp(s,getString())!=0);
	}

bool	Api::operator != (const Api& s) const throw(){
	return (strcmp(s.getString(),getString())!=0);
	}

bool	Api::operator < (const Api& s) const throw(){
	return (strcmp(s.getString(),getString())<0)?true:false;
	}

void	Api::fillFromBuffer(const void* buffer,unsigned long nChars) throw(){
	unsigned long	maxLen	= maxLength();
	unsigned long	n;
	if(nChars > maxLen){
		n			= maxLen;
		_truncated	= true;
		}
	else{
		n			= nChars;
		_truncated	= false;
		}
	char*			p		= getBuffer();
	memcpy(p,(const char*)buffer,n);
	p[n]	= '\0';	// This is cleared here to enable fast truncation check.
	}

Api::operator const char* () const throw(){
	return getString();
	}

unsigned long Api::length() const throw(){
	return strlen(getString());
	}

