/*
   Copyright (C) 2004 Michael Nelson Moran

   This file is part of OSCL.

   OSCL is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   OSCL is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   You should have received a copy of the GNU General Public License
   along with OSCL; see the file COPYING.  If not, write to the Free
   Software Foundation, 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To negotiate other licensing arrangments/agreements contact
   the copyright holder:

   Michael N. Moran
   5009 Old Field Ct.         mike@mnmoran.org
   Kennesaw, GA, USA 30144    http://mnmoran.org
*/

#include "var.h"

namespace Oscl {
namespace TriState {

class VarIsAsserted : public TriState::Var::Query {
	public:
		bool	_isAsserted;
	public:
		inline void asserted() throw(){
			_isAsserted	= true;
			}
		inline void negated() throw(){
			_isAsserted	= false;
			}
		inline void inhibited() throw(){
			_isAsserted	= false;
			}
	};

class VarIsNegated : public TriState::Var::Query {
	public:
		bool	_isNegated;
	public:
		inline void asserted() throw(){
			_isNegated	= false;
			}
		inline void negated() throw(){
			_isNegated	= true;
			}
		inline void inhibited() throw(){
			_isNegated	= false;
			}
	};

class IsInhibited : public TriState::Var::Query {
	public:
		bool	_isInhibited;
	public:
		inline void asserted() throw(){
			_isInhibited	= false;
			}
		inline void negated() throw(){
			_isInhibited	= false;
			}
		inline void inhibited() throw(){
			_isInhibited	= true;
			}
	};

class VarAsserted : public TriState::Var::State {
	public:
		bool operator ==(const TriState::Var::State& other) const throw();
		bool operator !=(const TriState::Var::State& other) const throw();
		void accept(TriState::Var::Query& q) const throw();
	};

class VarNegated : public TriState::Var::State {
	public:
		bool operator ==(const TriState::Var::State& other) const throw();
		bool operator !=(const TriState::Var::State& other) const throw();
		void accept(TriState::Var::Query& q) const throw();
	};

class VarInhibited : public TriState::Var::State {
	public:
		bool operator ==(const TriState::Var::State& other) const throw();
		bool operator !=(const TriState::Var::State& other) const throw();
		void accept(TriState::Var::Query& q) const throw();
	};

};
};

using namespace Oscl;

static const TriState::VarAsserted	_asserted;
static const TriState::VarNegated		_negated;
static const TriState::VarInhibited	_inhibited;

bool TriState::VarAsserted::operator ==(const TriState::Var::State& other) const throw(){
	TriState::VarIsAsserted	assertedQuery;
	other.accept(assertedQuery);
	return assertedQuery._isAsserted;
	}

bool TriState::VarAsserted::operator !=(const TriState::Var::State& other) const throw(){
	TriState::VarIsAsserted	assertedQuery;
	other.accept(assertedQuery);
	return !assertedQuery._isAsserted;
	}

void TriState::VarAsserted::accept(TriState::Var::Query& q) const throw(){
	q.asserted();
	}

bool TriState::VarNegated::operator ==(const TriState::Var::State& other) const throw(){
	TriState::VarIsNegated	negatedQuery;
	other.accept(negatedQuery);
	return negatedQuery._isNegated;
	}

bool TriState::VarNegated::operator !=(const TriState::Var::State& other) const throw(){
	TriState::VarIsNegated	negatedQuery;
	other.accept(negatedQuery);
	return !negatedQuery._isNegated;
	}

void TriState::VarNegated::accept(TriState::Var::Query& q) const throw(){
	q.negated();
	}

bool TriState::VarInhibited::operator ==(const TriState::Var::State& other) const throw(){
	IsInhibited	inhbitedQuery;
	other.accept(inhbitedQuery);
	return inhbitedQuery._isInhibited;
	}

bool TriState::VarInhibited::operator !=(const TriState::Var::State& other) const throw(){
	IsInhibited	inhbitedQuery;
	other.accept(inhbitedQuery);
	return !inhbitedQuery._isInhibited;
	}

void TriState::VarInhibited::accept(TriState::Var::Query& q) const throw(){
	q.inhibited();
	}

void	TriState::Var::accept(Query& q) const throw(){
	_state->accept(q);
	}

class GetState : public TriState::Var::Query {
	public:
		const TriState::Var::State*	_state;
	public:
		inline void asserted() throw(){
			_state	= &TriState::Var::getAsserted();
			}
		inline void negated() throw(){
			_state	= &TriState::Var::getNegated();
			}
		inline void inhibited() throw(){
			_state	= &TriState::Var::getInhibited();
			}
	};

TriState::Var::Var(const State& initial) throw():
		_state(&initial)
		{
	}

bool	TriState::Var::operator == (const TriState::Var& other) throw(){
	return *other._state == *_state;
	}

bool	TriState::Var::operator != (const TriState::Var& other) throw(){
	return *other._state != *_state;
	}

void	TriState::Var::operator=(const TriState::Var& newState) throw(){
	*this	= *newState._state;
	}

void	TriState::Var::operator=(const State& newState) throw(){
	GetState	getState;
	newState.accept(getState);
	_state	= getState._state;
	}

const TriState::Var::State&	TriState::Var::getCurrentState() const throw(){
	return *_state;
	}

void	TriState::Var::assert() throw(){
	_state	= &_asserted;
	}

void	TriState::Var::negate() throw(){
	_state	= &_negated;
	}

void	TriState::Var::inhibit() throw(){
	_state	= &_inhibited;
	}

const TriState::Var::State&	TriState::Var::getAsserted() throw(){
	return _asserted;
	}

const TriState::Var::State&	TriState::Var::getNegated() throw(){
	return _negated;
	}

const TriState::Var::State&	TriState::Var::getInhibited() throw(){
	return _inhibited;
	}

